# Generated by Django 2.2.13 on 2021-03-19 10:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0034_task_region"),
    ]

    operations = [
        migrations.AddField(
            model_name="workflow",
            name="guidelines_url",
            field=models.TextField(null=True),
        ),
        migrations.AddField(
            model_name="workflow",
            name="is_running",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="workflow",
            name="task_description",
            field=models.TextField(null=True),
        ),
    ]



import ctypes
import logging
import re

from django.utils import timezone
from email_validator import (
    EmailNotValidError,
    EmailSyntaxError,
    validate_email,
)
from rest_framework.exceptions import ErrorDetail
from rest_framework.views import exception_handler

logger = logging.getLogger(__name__)


def process_errors(error):
    error_list = []
    if isinstance(error, list):
        for message in error:
            error_list.append({"message": message})
    else:
        detail = error.get("detail")
        if detail:
            error_list.append({"message": detail})
            messages = error.get("messages", [])
            for message in messages:
                error_list.append({"message": message})
        else:
            fields = error.keys()
            for field in fields:
                if not isinstance(field, ErrorDetail) and isinstance(
                    error[field], dict
                ):
                    inner_fields = error[field].keys()
                    for inner_field in inner_fields:
                        for message in error[field][inner_field]:
                            error_object = {
                                "field": f"{field}-->{inner_field}",
                                "message": message,
                            }
                            error_list.append(error_object)
                else:
                    for message in error[field]:
                        error_object = {"field": field, "message": message}
                        error_list.append(error_object)
    return error_list


def custom_exception_handler(exc, context):
    # Call REST framework's default exception handler first,
    # to get the standard error response.
    response = exception_handler(exc, context)

    # Now add the HTTP status code to the response.
    if response is not None:
        response.data = {
            "status_code": response.status_code,
            "errors": process_errors(response.data),
        }
    return response


def is_invalid_email(email):
    try:
        validate_email(email)
        return False
    except (EmailSyntaxError, EmailNotValidError) as e:
        logger.info(f"{e}")
        return True


def generate_unique_token(*args):
    to_hash = str(timezone.now())
    for arg in args:
        to_hash += str(arg)
    token = ctypes.c_size_t(hash(to_hash)).value
    return token


def is_valid_url(text):
    if re.match(
        r"https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]"
        r"{1,6}\b([-a-zA-Z0-9()@:%_\+.~#?&//=]*)",
        text,
    ):
        return True
    return False



# Generated by Django 2.2.10 on 2020-04-10 14:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("user_handler", "0009_forgottenpassword"),
    ]

    operations = [
        migrations.AddField(
            model_name="invitation",
            name="admin",
            field=models.BooleanField(default=False),
        ),
    ]



from typing import Any, Dict, List

from django.core.management.base import BaseCommand, CommandParser

from human_lambdas.user_handler.models import Organization, User
from human_lambdas.workflow_handler.utils import STAFF_ORG_ID


class Command(BaseCommand):
    def add_arguments(self, parser: CommandParser):
        parser.add_argument("emails", nargs="+", type=str)

    def handle(self, *_: List[str], **options: Dict[str, Any]):
        if "emails" not in options:
            self.stdout.write(
                self.style.ERROR("invalid inputs, expected 1 or more email addresses")
            )
            return

        if not Organization.objects.filter(pk=STAFF_ORG_ID).exists():
            Organization(id=STAFF_ORG_ID, name="staff").save()

        templates_org = Organization.objects.get(pk=STAFF_ORG_ID)

        for email in options["emails"]:
            user = User.objects.get(email=email)
            templates_org.add_admin(user)

        self.stdout.write(
            self.style.SUCCESS(
                f"Added super admins {options['emails']} to org {STAFF_ORG_ID}"
            )
        )



import logging
from uuid import uuid4

from django.conf import settings
from django.db.models import Avg, F, Q
from django.utils import timezone
from drf_yasg2.utils import swagger_auto_schema
from rest_framework import serializers
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from human_lambdas.user_handler.models import Organization, User
from human_lambdas.user_handler.permissions import IsOrgAdmin
from human_lambdas.workflow_handler.models import Task, Workflow

logger = logging.getLogger(__name__)

MONTHS_BACK = 12
WEEKS_BACK = 14
DAYS_BACK = 28


def process_kwargs(**kwargs):
    query = Q(workflow__organization=kwargs["organization"]) & Q(
        workflow__disabled=False
    )
    if "workflow" in kwargs:
        query = query & Q(workflow=kwargs["workflow"])
    if "worker" in kwargs:
        query = query & Q(assigned_to=kwargs["worker"])
    return query


def get_completed(**kwargs):
    basic_query = process_kwargs(**kwargs)
    tasks = Task.objects.defer("data").filter(
        basic_query
        & Q(status="completed")
        & Q(completed_at__range=[kwargs["start_time"], kwargs["end_time"]])
    )
    return tasks.count()


def get_pending(**kwargs):
    basic_query = process_kwargs(**kwargs)
    result = (
        Task.objects.defer("data")
        .filter(
            basic_query
            & Q(created_at__lte=kwargs["end_time"])
            & Q(Q(completed_at__gt=kwargs["end_time"]) | Q(completed_at=None))
        )
        .count()
    )
    return result


def get_aht(**kwargs):
    basic_query = process_kwargs(**kwargs)
    result = (
        Task.objects.defer("data")
        .filter(
            basic_query
            & Q(status="completed")
            & Q(completed_at__range=[kwargs["start_time"], kwargs["end_time"]])
        )
        .aggregate(aht=Avg(F("handling_time_seconds")))
    )
    aht = result["aht"]
    return aht if aht else None


def get_tat(**kwargs):
    basic_query = process_kwargs(**kwargs)
    result = (
        Task.objects.defer("data")
        .filter(
            basic_query
            & Q(status="completed")
            & Q(completed_at__range=[kwargs["start_time"], kwargs["end_time"]])
        )
        .aggregate(tat=Avg(F("completed_at") - F("created_at")))
    )
    tat = result["tat"]
    return tat / timezone.timedelta(seconds=1) if tat else None


def get_accuracy(**kwargs):
    basic_query = process_kwargs(**kwargs)
    all_audited = Task.objects.defer("data").filter(
        basic_query
        & ~Q(correct=None)
        & Q(completed_at__range=[kwargs["start_time"], kwargs["end_time"]])
    )

    if not all_audited.exists():
        return None

    num_audited = float(len(all_audited))
    num_correct = float(len(all_audited.filter(Q(correct=True))))

    return num_correct / num_audited


METRICS = {
    "completed": get_completed,
    "pending": get_pending,
    "aht": get_aht,
    "tat": get_tat,
    "accuracy": get_accuracy,
}

WORKER_METRICS = {
    "completed": get_completed,
    "aht": get_aht,
    "accuracy": get_accuracy,
}


class WorkflowMetricsQuerySerializer(serializers.Serializer):
    range = serializers.CharField(required=False)
    type = serializers.MultipleChoiceField(list(METRICS.keys()), required=False)


def process_monthly():
    end = timezone.now()
    start = end.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
    time_ranges = [(start, end)]
    for _ in range(MONTHS_BACK):
        end = start
        start = (end - timezone.timedelta(days=1)).replace(day=1)
        time_ranges.append((start, end))
    return time_ranges


def process_weekly():
    end = timezone.now()
    start = (end - timezone.timedelta(days=end.isoweekday() - 1)).replace(
        hour=0, minute=0, second=0, microsecond=0
    )
    time_ranges = [(start, end)]
    for _ in range(WEEKS_BACK):
        end = start
        start = end - timezone.timedelta(days=7)
        time_ranges.append((start, end))
    return time_ranges


def process_daily():
    end = timezone.now()
    start = end.replace(hour=0, minute=0, second=0, microsecond=0)
    time_ranges = [(start, end)]
    for _ in range(DAYS_BACK):
        end = start
        start = end - timezone.timedelta(days=1)
        time_ranges.append((start, end))
    return time_ranges


_TIME_RANGE_DICT = {
    "daily": process_daily,
    "weekly": process_weekly,
    "monthly": process_monthly,
}


class OrganizationMetrics(APIView):
    permission_classes = (IsAuthenticated, IsOrgAdmin)

    def get_queryset(self):
        return Organization.objects.filter(pk=self.kwargs["org_id"])

    def validate_data(self, data):
        pass

    def process_time_range(self, range_name):
        return _TIME_RANGE_DICT[range_name]()

    @swagger_auto_schema(query_serializer=WorkflowMetricsQuerySerializer)
    def get(self, request, *args, **kwargs):
        if settings.DATABASES["default"]["ENGINE"] == "django.db.backends.sqlite3":
            return Response(None, status=204)

        self.validate_data(request.data)
        data = []
        qtypes = request.query_params.getlist("type")
        if any([qtype in METRICS for qtype in qtypes]):
            organization = self.get_queryset().first()
            time_ranges = self.process_time_range(request.query_params.get("range"))
            for start_time, end_time in reversed(time_ranges):
                data_dict = {
                    "date": (end_time - timezone.timedelta(microseconds=1)).replace(
                        tzinfo=None
                    ),
                    "id": uuid4().hex,
                }
                for qtype in qtypes:
                    try:
                        data_dict[qtype] = METRICS[qtype](
                            start_time=start_time,
                            end_time=end_time,
                            organization=organization,
                        )
                    except KeyError:
                        continue
                data.append(data_dict)
        return Response(data, status=200)


class WorkflowMetrics(APIView):
    permission_classes = (IsAuthenticated, IsOrgAdmin)

    def get_queryset(self):
        return Organization.objects.filter(pk=self.kwargs["org_id"])

    def validate_data(self, data):
        pass

    def process_time_range(self, range_name):
        return _TIME_RANGE_DICT[range_name]()

    def get(self, request, *args, **kwargs):
        self.validate_data(request.data)
        qtypes = request.query_params.getlist("type")
        workflow_ids = request.query_params.getlist("workflow_id")
        if any([qtype in METRICS for qtype in qtypes]):
            organization = self.get_queryset().first()
            if workflow_ids:
                workflows = Workflow.objects.filter(
                    organization=organization, pk__in=workflow_ids, disabled=False
                ).all()
            else:
                workflows = Workflow.objects.filter(
                    organization=organization, disabled=False
                ).all()
            time_ranges = self.process_time_range(request.query_params.get("range"))
            result = {}
            for qtype in qtypes:
                data = []
                for start_time, end_time in reversed(time_ranges):
                    data_dict = {
                        "date": end_time,
                        "id": uuid4().hex,
                    }

                    for workflow in workflows:
                        try:
                            data_dict[workflow.name] = METRICS[qtype](
                                start_time=start_time,
                                end_time=end_time,
                                organization=organization,
                                workflow=workflow,
                            )
                        except KeyError:
                            continue
                    data.append(data_dict)
                result[qtype] = data
        return Response(result, status=200)


class WorkerMetrics(APIView):
    permission_classes = (IsAuthenticated, IsOrgAdmin)

    def get_queryset(self):
        return Organization.objects.filter(pk=self.kwargs["org_id"])

    def validate_data(self, data):
        pass

    def process_time_range(self, range_name):
        return _TIME_RANGE_DICT[range_name]()

    def get(self, request, *args, **kwargs):
        self.validate_data(request.data)
        qtypes = request.query_params.getlist("type")
        worker_id = request.query_params.getlist("worker_id")
        if any([qtype in WORKER_METRICS for qtype in qtypes]):
            organization = self.get_queryset().first()
            if worker_id:
                users = User.objects.filter(
                    organization=organization,
                    pk__in=worker_id,
                ).all()
            else:
                users = User.objects.filter(organization=organization).all()
            time_ranges = self.process_time_range(request.query_params.get("range"))
            result = {}
            for qtype in qtypes:
                data = []
                for start_time, end_time in reversed(time_ranges):
                    data_dict = {
                        "date": end_time,
                        "id": uuid4().hex,
                    }

                    for user in users:
                        try:
                            data_dict[user.name] = WORKER_METRICS[qtype](
                                start_time=start_time,
                                end_time=end_time,
                                organization=organization,
                                worker=user,
                            )
                        except KeyError:
                            continue
                    data.append(data_dict)
                result[qtype] = data
        return Response(result, status=200)



import analytics
from django.apps import AppConfig
from django.conf import settings


class WorkflowHandlerConfig(AppConfig):
    name = "human_lambdas.workflow_handler"

    def ready(self):
        analytics.write_key = settings.SEGMENT_KEY



from django.conf import settings
from google.auth.transport import requests
from google.oauth2 import id_token
from rest_framework import serializers
from rest_framework.generics import CreateAPIView
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework_simplejwt.exceptions import TokenError
from rest_framework_simplejwt.serializers import TokenObtainPairSerializer
from rest_framework_simplejwt.tokens import RefreshToken
from rest_framework_simplejwt.views import TokenObtainPairView

from human_lambdas.hl_rest_api import analytics
from human_lambdas.user_handler.models import User


class HLTokenObtainPairSerializer(TokenObtainPairSerializer):
    def is_valid(self, raise_exception):
        user = User.objects.filter(email=self.initial_data["email"]).first()

        if user and len(user.password) == 0:
            raise TokenError("This account only uses Google sign in")

        return super().is_valid(raise_exception=raise_exception)

    @classmethod
    def get_token(cls, user):
        token = super().get_token(user)
        token["email"] = user.email
        token["name"] = user.name
        analytics.track(user.pk, "Get Token")
        return token


class HLTokenObtainPairView(TokenObtainPairView):
    serializer_class = HLTokenObtainPairSerializer


class GoogleTokenSerializer(serializers.Serializer):
    token = serializers.CharField()


class GoogleToken(CreateAPIView):
    serializer_class = GoogleTokenSerializer
    permission_classes = (AllowAny,)

    def create(self, request):
        serializer = GoogleTokenSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        token = request.data["token"]
        try:
            idinfo = id_token.verify_oauth2_token(
                token, requests.Request(), settings.GOOGLE_OAUTH_CLIENT_ID
            )
        except ValueError as ex:
            return Response(
                {
                    "status_code": 400,
                    "errors": [
                        {"message": f"Invalid token. {ex.args[0]}\n\n'{token}'"}
                    ],
                },
                status=400,
            )
        email = idinfo["email"]
        users = User.objects.filter(email=email)
        if not users.exists():
            return Response(
                {
                    "status_code": 400,
                    "errors": [{"message": f"no user exists for {email}"}],
                },
                status=400,
            )

        assert len(users) == 1
        refresh = RefreshToken.for_user(users.first())

        return Response(
            {
                "refresh": str(refresh),
                "access": str(refresh.access_token),
            },
            status=200,
        )



import copy
import logging
import os

from rest_framework import status

from human_lambdas.user_handler.models import Organization
from human_lambdas.workflow_handler.models import Task, Workflow
from human_lambdas.workflow_handler.tests.constants import (
    REGISTRATION_DATA,
    REGISTRATION_DATA_2,
    WORKFLOW_DATA_3,
)
from human_lambdas.workflow_handler.tests.util import HLTestCase
from human_lambdas.workflow_handler.utils import STAFF_ORG_ID

logger = logging.getLogger(__name__)

_CURRENT_DIR = os.path.dirname(__file__)


class TestTasksSave(HLTestCase):
    def setUserClient(self, email):
        response = self.client.post(
            "/v1/users/token", {"email": email, "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)

    def setUp(self):
        self.file_path = os.path.join(_CURRENT_DIR, "data", "test.csv")

        response = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.user_id = response.data["id"]

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA_2)

        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)

        _ = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            WORKFLOW_DATA_3,
            format="json",
        )
        self.workflow_id = Workflow.objects.get(name="uploader").id
        with open(self.file_path) as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(
                    self.org_id, self.workflow_id
                ),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

    def test_save_single_selection_task(self):
        workflow = Workflow.objects.get(id=self.workflow_id)
        tasks = Task.objects.filter(workflow=workflow).all()
        output_values = ["foo1", "bar1"]
        for output_value, task in zip(output_values, tasks):

            _ = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/tasks/{2}/assign".format(
                    self.org_id, self.workflow_id, task.id
                ),
                data={"assigned_to": self.user_id},
                format="json",
            )

            data = copy.deepcopy(task.data)
            for idata in data:
                if idata["id"] == "foo":
                    idata["single_selection"]["value"] = output_value
            data_payload = {"data": data}

            response = self.client.patch(
                "/v1/orgs/{0}/workflows/{1}/tasks/{2}/save".format(
                    self.org_id, self.workflow_id, task.id
                ),
                data=data_payload,
                format="json",
            )

            self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)
            self.assertEqual("in_progress", response.data["status"])
            self.assertEqual(data, response.data["data"])
            self.assertEqual("in_progress", Task.objects.get(id=task.id).status)
            self.assertIsNone(Task.objects.get(id=task.id).completed_at)

            # test retrieving the same task
            response = self.client.get(
                "/v1/orgs/{0}/workflows/{1}/tasks/{2}".format(
                    self.org_id, self.workflow_id, task.id
                )
            )
            self.assertEqual("in_progress", response.data["status"])
            self.assertEqual(data, response.data["data"])

    def test_when_worker_tries_to_save_task_then_passes_auth(self):
        workflow = Workflow.objects.get(id=self.workflow_id)
        task = Task.objects.filter(workflow=workflow).first()
        org = Organization.objects.get(id=self.org_id)
        org.admin.set([])
        org.save()
        response = self.client.patch(
            f"/v1/orgs/{self.org_id}/workflows/{self.workflow_id}/tasks/{task.id}/save",
            data=task.data[0],
            format="json",
        )

        assert (
            response.status_code == status.HTTP_400_BAD_REQUEST
        )  # complains not assigned == good

    def test_when_worker_tries_to_save_task_in_admin_org_then_error(self):
        workflow = Workflow.objects.get(id=self.workflow_id)
        task = Task.objects.filter(workflow=workflow).first()
        response = self.client.patch(
            f"/v1/orgs/{STAFF_ORG_ID}/workflows/{self.workflow_id}/tasks/{task.id}/save",
            data=task.data[0],
            format="json",
        )

        assert response.status_code == status.HTTP_403_FORBIDDEN



# Generated by Django 2.2.5 on 2020-01-13 13:14

from django.db import migrations

import human_lambdas.workflow_handler.fields


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0009_auto_20200109_1621"),
    ]

    operations = [
        migrations.AlterField(
            model_name="task",
            name="outputs",
            field=human_lambdas.workflow_handler.fields.JSONField(default={}),
            preserve_default=False,
        ),
    ]



"""
Django settings for hl-rest-api project.

Generated by 'django-admin startproject' using Django 2.2.5.

For more information on this file, see
https://docs.djangoproject.com/en/2.2/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.2/ref/settings/
"""

import os
import sys
import uuid
from datetime import timedelta
from pathlib import Path

import requests
import sentry_sdk
from sentry_sdk.integrations.django import DjangoIntegration

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = os.getenv("SECRET_KEY")

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = False if os.getenv("DEBUG") == "False" else True


HOOK_EVENTS = {
    "task.completed": "workflow_handler.Task.completed",
}
HOOK_CUSTOM_MODEL = "workflow_handler.WebHook"
HOOK_FINDER = "human_lambdas.workflow_handler.utils.find_and_fire_hook"

# Application definition

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "human_lambdas.user_handler",
    "human_lambdas.workflow_handler",
    "human_lambdas.templates_handler",
    "human_lambdas.external",
    "rest_framework",
    "rest_framework_simplejwt",
    "corsheaders",
    # "rest_framework.authtoken",
    "rest_hooks",
    "drf_yasg2",
]

APPEND_SLASH = False

SENDGRID_API_KEY = os.getenv("SENDGRID_API_KEY")

HL_CLIENT_API_KEY = os.getenv("HL_CLIENT_API_KEY")
HL_ORG_ID = os.getenv("HL_ORG_ID")

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "human_lambdas.middleware.RewriteMiddleware",
    "whitenoise.middleware.WhiteNoiseMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "corsheaders.middleware.CorsMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "human_lambdas.hl_rest_api.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "human_lambdas.hl_rest_api.wsgi.application"

if "POSTGRES_DB" in os.environ:
    print("Running against Postgres")
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.postgresql_psycopg2",
            "NAME": os.getenv("POSTGRES_DB"),
            "USER": os.getenv("POSTGRES_USER"),
            "PASSWORD": os.getenv("POSTGRES_PASSWORD"),
            "HOST": os.getenv("POSTGRES_HOST"),
            "PORT": os.getenv("POSTGRES_PORT"),
        }
    }
else:
    print("No POSTGRES_DB given, using sqlite")
    db_path = Path.cwd() / ".human_lambdas"
    db_path.mkdir(parents=True, exist_ok=True)
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.sqlite3",
            "NAME": (db_path / "db").as_posix(),
        },
    }


# Password validation
# https://docs.djangoproject.com/en/2.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
]


# Internationalization
# https://docs.djangoproject.com/en/2.2/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.2/howto/static-files/

STATIC_URL = "/"
STATIC_ROOT = os.path.join(BASE_DIR, "html", "build")
WHITENOISE_INDEX_FILE = True

AUTH_USER_MODEL = "user_handler.User"

if DEBUG:
    DEFAULT_PERMISSION = "rest_framework.permissions.AllowAny"
    ALLOWED_HOSTS = ["*"]
else:
    ALLOWED_HOSTS = [
        # "human-lambdas-api.eu-west-2.elasticbeanstalk.com",
        # "human-lambdas-api-stg.eu-west-2.elasticbeanstalk.com",
        # "human-lambdas-api-redesign.eu-west-2.elasticbeanstalk.com",
        # "api.humanlambdas.com",
        # "api-stg.humanlambdas.com",
        # "localhost",
        "*",
    ]
    DEFAULT_PERMISSION = "rest_framework.permissions.IsAuthenticated"

REST_FRAMEWORK = {
    # Use Django's standard `django.contrib.auth` permissions,
    # or allow read_only access for unauthenticated users.
    "DEFAULT_PERMISSION_CLASSES": [DEFAULT_PERMISSION],
    "DEFAULT_AUTHENTICATION_CLASSES": [
        "rest_framework_simplejwt.authentication.JWTAuthentication",
    ],
    "EXCEPTION_HANDLER": "human_lambdas.hl_rest_api.utils.custom_exception_handler",
    # "DEFAULT_SCHEMA_CLASS": "rest_framework.schemas.coreapi.AutoSchema",
}

ACCOUNT_USER_MODEL_USERNAME_FIELD = None
ACCOUNT_AUTHENTICATION_METHOD = "email"

ACCOUNT_EMAIL_REQUIRED = True
ACCOUNT_UNIQUE_EMAIL = True
ACCOUNT_USERNAME_REQUIRED = False
ACCOUNT_USER_EMAIL_FIELD = "email"
ACCOUNT_LOGOUT_ON_GET = True

CORS_ORIGIN_ALLOW_ALL = True
CORS_ALLOW_CREDENTIALS = True

REST_AUTH_SERIALIZERS = {
    "USER_DETAILS_SERIALIZER": "human_lambdas.user_handler.serializers.UserLoginSerializer",
}
REST_AUTH_REGISTER_SERIALIZERS = {
    "REGISTER_SERIALIZER": "human_lambdas.user_handler.serializers.UserCreateSerializer",
}

REGIONAL_BUCKET_AU = os.getenv("REGIONAL_BUCKET_AU")
REGIONAL_BUCKET_US = os.getenv("REGIONAL_BUCKET_US")
DEFAULT_DEV_BUCKET = "data-local-dev"

TESTING = "test" in sys.argv
if TESTING:
    LOGGING = {}
else:
    LOGGING = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "verbose": {
                "format": "%(levelname)s %(asctime)s %(module)s %(process)d %(thread)d %(message)s"
            },
            "simple": {"format": "%(levelname)s %(asctime)s %(message)s"},
        },
        "handlers": {
            "console": {
                "level": "INFO",
                "class": "logging.StreamHandler",
                "formatter": "simple",
            },
        },
        "loggers": {
            "django": {
                "handlers": ["console"],
                "level": "INFO",
                "propagate": True,
            },
            "human_lambdas.workflow_handler": {
                "handlers": ["console"],
                "level": "INFO",
                "propagate": True,
            },
            "human_lambdas.user_handler": {
                "handlers": ["console"],
                "level": "INFO",
                "propagate": True,
            },
            "organization_handler": {
                "handlers": ["console"],
                "level": "INFO",
                "propagate": True,
            },
            "metrics": {
                "handlers": ["console"],
                "level": "INFO",
                "propagate": True,
            },
            "human_lambdas.external": {
                "handlers": ["console"],
                "level": "INFO",
                "propagate": True,
            },
        },
    }

SIMPLE_JWT = {
    # https://github.com/davesque/django-rest-framework-simplejwt
    "ACCESS_TOKEN_LIFETIME": timedelta(minutes=15),
    "ALGORITHM": "HS256",
    "REFRESH_TOKEN_LIFETIME": timedelta(days=14),
    "SIGNING_KEY": os.getenv("SECRET_KEY"),
    "ROTATE_REFRESH_TOKENS": True,
}


if not DEBUG:
    SENTRY_KEY = os.getenv("SENTRY_KEY")
    SENTRY_PROJECT = os.getenv("SENTRY_PROJECT")

    if not SENTRY_KEY is None and not SENTRY_PROJECT is None:
        sentry_sdk.init(
            dsn="https://{0}@sentry.io/{1}".format(SENTRY_KEY, SENTRY_PROJECT),
            integrations=[DjangoIntegration()],
            # If you wish to associate users to errors (assuming you are using
            # django.contrib.auth) you may enable sending PII data.
            send_default_pii=True,
        )

FRONT_END_BASE_URL = os.getenv("APP_URL", "http://localhost:3000/")

if not FRONT_END_BASE_URL.endswith("/"):
    FRONT_END_BASE_URL = f"{FRONT_END_BASE_URL}/"

API_URL = os.getenv("API_URL", "http://localhost:8000")

LOCAL_IP = None
try:
    LOCAL_IP = requests.get(
        "http://169.254.169.254/latest/meta-data/local-ipv4", timeout=0.01
    ).text
except requests.exceptions.RequestException:
    pass
if LOCAL_IP and not DEBUG:
    ALLOWED_HOSTS.append(LOCAL_IP)


SEGMENT_KEY = os.environ.get("SEGMENT_KEY")

# Notification email
SEND_NOTIFICATION_TEMPLATE = os.environ.get("SEND_NOTIFICATION_TEMPLATE")
NOTIFICATION_ASM_GROUPID = os.environ.get("NOTIFICATION_ASM_GROUPID")

THROTTLING_TIME_MIN = 5

# Account Emails
INVITATION_EXPIRATION_WINDOW_DAYS = 30
INVITATION_TEMPLATE = os.environ.get("INVITATION_TEMPLATE")
ACCOUNT_ASM_GROUPID = os.environ.get("ACCOUNT_ASM_GROUPID")
FORGOTTEN_PASSWORD_TEMPLATE = os.environ.get("FORGOTTEN_PASSWORD_TEMPLATE")


# Swagger
SWAGGER_SETTINGS = {
    "SECURITY_DEFINITIONS": {
        "Token": {"type": "apiKey", "name": "Authorization", "in": "header"},
    },
}

GOOGLE_OAUTH_CLIENT_ID = (
    "725636052597-sp46vdemr0u2s0hh698nqupjdvucoa3o.apps.googleusercontent.com"
)

# AU_DATA_BUCKET = os.environ["AU_DATA_BUCKET"]
# NA_DATA_BUCKET = os.environ["NA_DATA_BUCKET"]

if os.getenv("PYTEST") == "true":
    STORAGE_TEST_PREFIX = str(uuid.uuid4())
else:
    STORAGE_TEST_PREFIX = None

DEV_BUCKET = "data-local-dev"

REGIONAL_BUCKET_AU = os.getenv("REGIONAL_BUCKET_AU", DEV_BUCKET)
REGIONAL_BUCKET_US = os.getenv("REGIONAL_BUCKET_US", DEV_BUCKET)



import logging

from django.conf import settings
from django.shortcuts import get_object_or_404
from django.utils import timezone
from rest_framework import serializers
from rest_framework.generics import RetrieveUpdateAPIView
from rest_framework.permissions import IsAuthenticated

from human_lambdas.hl_rest_api import analytics

from .apps import UserHandlerConfig
from .models import Notification, User

logger = logging.getLogger(__name__)


class NotificationSerializer(serializers.ModelSerializer):
    queue_notifications = serializers.JSONField()

    class Meta:
        model = Notification
        fields = "__all__"
        extra_kwargs = {}

    def to_representation(self, instance):
        return instance.format_output()

    def update(self, instance, validated_data):
        instance.enabled = validated_data.get("enabled", instance.enabled)
        workflow_notifications = validated_data.get("queue_notifications", [])
        any_enabled = False
        for workflow_notification in workflow_notifications:
            wfnotification_obj = instance.workflownotification_set.filter(
                workflow__pk=workflow_notification["queue_id"],
                workflow__disabled=False,
            ).first()
            if wfnotification_obj:
                wfnotification_obj.enabled = workflow_notification["enabled"]
                wfnotification_obj.save()
                any_enabled = True if workflow_notification["enabled"] else False
        instance.save()
        analytics.track(
            self.context["request"].user.pk,
            "Notifications Update",
            {"all_enabled": instance.enabled, "any_enabled": any_enabled},
        )
        return instance


class NotificationView(RetrieveUpdateAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = NotificationSerializer

    def get_queryset(self):
        user = self.request.user
        return Notification.objects.filter(pk=user.notifications.pk)

    def get_object(self):
        obj = get_object_or_404(self.get_queryset())
        return obj


def send_notification(workflow):
    users = User.objects.filter(
        notifications__workflownotification__workflow=workflow,
        notifications__enabled=True,
    ).distinct()
    emails = []
    for user in users:
        wf_notification = user.notifications.workflownotification_set.filter(
            workflow=workflow
        ).first()
        if not wf_notification.enabled or (
            wf_notification.last_notified
            and timezone.now() - wf_notification.last_notified
            < timezone.timedelta(minutes=settings.THROTTLING_TIME_MIN)
        ):
            continue
        emails.append(user.email)
        wf_notification.last_notified = timezone.now()
        wf_notification.save()
    if emails:
        if (
            settings.SEND_NOTIFICATION_TEMPLATE is None
            or settings.NOTIFICATION_ASM_GROUPID is None
        ):
            logger.warn("Sendgrid not set up, ignoring email trigger")
            return

        template_data = {
            "workflow_name": workflow.name,
            "org_id": workflow.organization.pk,
            "hl_url": settings.FRONT_END_BASE_URL,
            "workflow_id": workflow.pk,
        }
        UserHandlerConfig.emailclient.send_email(
            to_email=emails,
            template_id=settings.SEND_NOTIFICATION_TEMPLATE,
            template_data=template_data,
            group_id=int(settings.NOTIFICATION_ASM_GROUPID),
        )



# Generated by Django 2.2.5 on 2020-02-06 14:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("user_handler", "0005_invitation"),
    ]

    operations = [
        migrations.AlterField(
            model_name="user",
            name="name",
            field=models.CharField(max_length=128, null=True),
        ),
    ]



from django.db.models import Q
from django.utils import timezone
from rest_framework import serializers
from rest_framework.generics import ListCreateAPIView, RetrieveDestroyAPIView
from rest_framework.permissions import IsAuthenticated

from .models import Organization, Task, TaskActivity, Workflow
from .serializers import TaskSerializer
from .utils import STAFF_ORG_ID, get_session_duration_seconds


class ActivitySerializer(serializers.ModelSerializer):
    assignee_name = serializers.SerializerMethodField(read_only=True)
    created_by_name = serializers.SerializerMethodField(read_only=True)

    class Meta:
        fields = "__all__"
        model = TaskActivity

    def get_assignee_name(self, obj):
        if obj.assignee:
            return obj.assignee.name
        return None

    def get_created_by_name(self, obj):
        return obj.created_by.name


def get_task(user, org_id, workflow_id, task_id):
    if org_id == STAFF_ORG_ID:
        return Task.objects.defer("data").filter(pk=task_id).first()

    organizations = Organization.objects.filter(user=user).all()
    workflows = Workflow.objects.filter(
        Q(organization__in=organizations) & Q(organization__pk=org_id)
    )
    return (
        Task.objects.defer("data")
        .filter(Q(workflow__in=workflows) & Q(workflow=workflow_id) & Q(pk=task_id))
        .first()
    )


class ActivityView(ListCreateAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = ActivitySerializer
    task_serializer_class = TaskSerializer

    def get_queryset(self):
        user = self.request.user

        task: Task = get_task(
            user,
            self.kwargs["org_id"],
            self.kwargs["workflow_id"],
            self.kwargs["task_id"],
        )

        if user.id == task.assigned_to_id:
            task.session_started_at = timezone.now()
            task.save()

        return TaskActivity.objects.filter(task=task).order_by("-created_at")

    def post(self, request, *args, **kwargs):
        request.data["created_by"] = self.request.user.pk
        request.data["task"] = self.kwargs["task_id"]

        user = self.request.user

        task = get_task(
            user,
            self.kwargs["org_id"],
            self.kwargs["workflow_id"],
            self.kwargs["task_id"],
        )

        if task.assigned_to and user.id == task.assigned_to.id:
            task.handling_time_seconds += get_session_duration_seconds(task)
            task.session_started_at = None
            task.save()

        return self.create(request, *args, **kwargs)


class RDActivityView(RetrieveDestroyAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = ActivitySerializer

    def get_queryset(self):
        user = self.request.user
        if getattr(self, "swagger_fake_view", False):
            # queryset just for schema generation metadata
            return TaskActivity.objects.none()

        task = get_task(
            user,
            self.kwargs["org_id"],
            self.kwargs["workflow_id"],
            self.kwargs["task_id"],
        )

        return TaskActivity.objects.filter(task=task, action="comment")

    def delete(self, request, *args, **kwargs):
        user = self.request.user
        task = get_task(
            user,
            self.kwargs["org_id"],
            self.kwargs["workflow_id"],
            self.kwargs["task_id"],
        )

        if task.assigned_to and user.id == task.assigned_to.id:
            task.handling_time_seconds += get_session_duration_seconds(task)
            task.session_started_at = None
            task.save()

        return self.destroy(request, *args, **kwargs)



from django.urls import path

from human_lambdas.metrics.views import (
    OrganizationMetrics,
    WorkerMetrics,
    WorkflowMetrics,
)

urlpatterns = [
    path("", OrganizationMetrics.as_view(), name="org-metrics"),
    path(
        "/workflows",
        WorkflowMetrics.as_view(),
        name="workflow-metrics",
    ),
    path(
        "/queues",
        WorkflowMetrics.as_view(),
        name="workflow-metrics",
    ),
    path("/workers", WorkerMetrics.as_view(), name="worker-metrics"),
]



from django.urls import include, path

urlpatterns = [
    path("/users", include("human_lambdas.v1.users")),
    path("/orgs", include("human_lambdas.v1.orgs")),
    path("/templates", include("human_lambdas.v1.templates")),
]



import csv
import logging
import os
from io import StringIO

import pytest
from django.test import TestCase
from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.data_handler.csv_utils import process_csv, validate_keys
from human_lambdas.user_handler.models import Organization, User
from human_lambdas.workflow_handler.models import Task, Workflow
from human_lambdas.workflow_handler.tests.constants import (
    ALPHA,
    BETA,
    GAMMA,
    REGISTRATION_DATA,
)

logger = logging.getLogger(__name__)

_CURRENT_DIR = os.path.dirname(__file__)


class TestUpload(APITestCase):
    def setUp(self):
        self.file_path = os.path.join(_CURRENT_DIR, "data", "test.csv")
        self.larger_file_path = os.path.join(_CURRENT_DIR, "data", "dataset.csv")

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.refresh = response.data["refresh"]

    def test_upload(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        workflow_data = {
            "name": "uploader",
            "data": [
                ALPHA,
                BETA,
                GAMMA,
                {
                    "id": "foo",
                    "name": "foo",
                    "type": "single_selection",
                    "single_selection": {"options": ["foo1", "bar1"]},
                },
            ],
        }
        _ = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        workflow_id = Workflow.objects.get(name="uploader").id
        with open(self.file_path, encoding="ISO-8859-1") as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(self.org_id, workflow_id),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

        task = Task.objects.filter(workflow__pk=workflow_id).first()
        self.assertIn(task.source.name, self.file_path)

    def test_larger_upload(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        workflow_data = {
            "name": "uploader",
            "data": [
                {
                    "id": "news",
                    "name": "news",
                    "type": "text",
                    "text": {"read_only": True},
                },
                {
                    "id": "type",
                    "name": "type",
                    "type": "text",
                    "text": {"read_only": True},
                },
                {
                    "id": "foo",
                    "name": "foo",
                    "type": "single_selection",
                    "single_selection": {"options": ["foo1", "bar1"]},
                },
            ],
        }
        _ = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        workflow_id = Workflow.objects.get(name="uploader").id
        with open(self.larger_file_path, encoding="ISO-8859-1") as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(self.org_id, workflow_id),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

    def test_task_creation(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        workflow_data = {
            "name": "uploader",
            "data": [
                ALPHA,
                BETA,
                GAMMA,
                {
                    "id": "foo",
                    "name": "foo",
                    "type": "single_selection",
                    "single_selection": {"options": ["foo1", "bar1"]},
                },
            ],
        }
        _ = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        workflow_id = Workflow.objects.get(name="uploader").id
        with open(self.file_path) as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(self.org_id, workflow_id),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)
        tasks = Task.objects.all()
        for task in tasks:
            self.assertEqual(4, len(task.data))
            for data_item in task.data:
                if data_item[data_item["type"]].get("read_only"):
                    self.assertTrue(data_item[data_item["type"]].pop("value"))
                    self.assertTrue(
                        any(wf_data == data_item for wf_data in workflow_data["data"])
                    )


class TestCSV2Task(TestCase):
    def setUp(self):
        self.test_csv_string = """alpha,beta,gamma,delta
        1,2,3,4
        5,6,7,8"""

        self.test_csv_string_missing_columns = """alpha,beta,delta
        1,2,4
        5,6,8"""

        self.test_csv_string_extra_columns = """alpha,beta,gamma,delta,lambda
        1,2,3,4,5
        6,7,8,9,10"""

        self.test_csv_string_duplicate_columns = """alpha,beta,gamma,delta,alpha
        1,2,3,4,1
        5,6,7,8,5"""

        self.test_csv_file = StringIO(self.test_csv_string)
        self.test_csv_file_missing_columns = StringIO(
            self.test_csv_string_missing_columns
        )
        self.test_csv_file_extra_columns = StringIO(self.test_csv_string_extra_columns)
        self.test_csv_file_duplicate_columns = StringIO(
            self.test_csv_string_duplicate_columns
        )
        user = User(name="foo", email="foo@bar.com")
        user.set_password("123")
        user.save()
        org = Organization(name="fooInc")
        org.save()
        org.add_admin(user)
        self.sample_workflow = Workflow(
            name="example",
            inputs=[
                {"id": "alpha", "name": "alpha", "type": "text"},
                {"id": "beta", "name": "beta", "type": "text"},
                {"id": "gamma", "name": "gamma", "type": "text"},
                {"id": "delta", "name": "delta", "type": "text"},
            ],
            outputs=[{"id": "binary", "name": "binary", "type": "binary"}],
            organization=org,
            created_by=user,
        )
        self.sample_workflow.save()

    def test_validate_keys_missing_columns(self):
        title_row = next(csv.reader(self.test_csv_file_missing_columns))
        try:
            validate_keys(title_row, self.sample_workflow)
        except Exception as e:
            self.assertEqual(str(e), "The dataset is missing some columns")

    def test_validate_keys_duplicate_columns(self):
        title_row = next(csv.reader(self.test_csv_file_duplicate_columns))
        try:
            validate_keys(title_row, self.sample_workflow)
        except Exception as e:
            self.assertEqual(str(e), "There are duplicate column names")

    def test_validate_keys_correct_columns(self):
        title_row = next(csv.reader(self.test_csv_file))
        try:
            validate_keys(title_row, self.sample_workflow)
        except Exception as e:
            self.fail("An error has occurred " + e)

    def test_validate_keys_extra_columns(self):
        title_row = next(csv.reader(self.test_csv_file_extra_columns))
        try:
            validate_keys(title_row, self.sample_workflow)
        except Exception as e:
            self.fail("An error has occurred " + e)

    def test_process_csv(self):
        user = User.objects.get(email="foo@bar.com")
        # try:
        process_csv(
            csv_file=self.test_csv_file,
            workflow=self.sample_workflow,
            source=None,
            filename="test.csv",
            user=user,
        )
        title_row = ["alpha", "beta", "gamma", "delta"]
        # Check each workflow key appears once in each task
        tasks = Task.objects.all()
        for task in tasks:
            assert task.region is None
            for input_item in task.inputs:
                self.assertEqual(1, title_row.count(input_item["id"]))


class TestUploadExtremes(APITestCase):
    def setUp(self):
        self.file_path = os.path.join(_CURRENT_DIR, "data", "data_space.csv")
        self.feedback_file_path = os.path.join(_CURRENT_DIR, "data", "feedback.csv")

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.refresh = response.data["refresh"]

    def test_upload_spaces(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        workflow_data = {
            "name": "uploader",
            "data": [
                {
                    "id": "car",
                    "name": "alpha",
                    "type": "text",
                    "text": {"read_only": True},
                },
                {
                    "id": "img_1",
                    "name": "beta",
                    "type": "text",
                    "text": {"read_only": True},
                },
                {
                    "id": "img_2",
                    "name": "gamma",
                    "type": "text",
                    "text": {"read_only": True},
                },
                {
                    "id": "foo",
                    "name": "foo",
                    "type": "single_selection",
                    "single_selection": {"options": ["foo1", "bar1"]},
                },
            ],
        }
        _ = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        workflow_id = Workflow.objects.get(name="uploader").id
        with open(self.file_path, encoding="ISO-8859-1") as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(self.org_id, workflow_id),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

    def test_upload_with_emoji(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        workflow_data = {
            "name": "feedback",
            "data": [
                {
                    "id": "submitter_picture",
                    "name": "alpha",
                    "type": "image",
                    "image": {"read_only": True},
                },
                {
                    "id": "submitter_full_name",
                    "name": "beta",
                    "type": "text",
                    "text": {"read_only": True},
                },
                {
                    "id": "submission_date",
                    "name": "gamma",
                    "type": "date",
                    "date": {"read_only": True},
                },
                {
                    "id": "subject",
                    "name": "eta",
                    "type": "text",
                    "text": {"read_only": True},
                },
                {
                    "id": "body",
                    "name": "leta",
                    "type": "text",
                    "text": {"read_only": True},
                },
                {
                    "id": "organization",
                    "name": "my",
                    "type": "text",
                    "text": {"read_only": True},
                },
                {
                    "id": "foo",
                    "name": "foo",
                    "type": "single_selection",
                    "single_selection": {"options": ["foo1", "bar1"]},
                },
            ],
        }
        _ = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        workflow_id = Workflow.objects.get(name="feedback").id
        with open(self.feedback_file_path) as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(self.org_id, workflow_id),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        workflow = Workflow.objects.get(pk=workflow_id)
        self.assertEqual(workflow.n_tasks, 6)


class TestUploadList(APITestCase):
    def setUp(self):
        self.file_path = os.path.join(_CURRENT_DIR, "data", "list_data.csv")

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.refresh = response.data["refresh"]

    def test_upload_list(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        workflow_data = {
            "name": "uploader",
            "data": [
                {
                    "id": "Alpha",
                    "name": "alpha",
                    "type": "list",
                    "list": {"read_only": True},
                    "subtype": "number",
                },
                {
                    "id": "Beta",
                    "name": "beta",
                    "type": "number",
                    "number": {"read_only": True},
                },
                {
                    "id": "Gamma",
                    "name": "gamma",
                    "type": "number",
                    "number": {"read_only": True},
                },
                {
                    "id": "foo",
                    "name": "foo",
                    "type": "single_selection",
                    "single_selection": {"options": ["foo1", "bar1"]},
                },
            ],
        }
        _ = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        workflow_id = Workflow.objects.get(name="uploader").id
        with open(self.file_path) as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(self.org_id, workflow_id),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

        for task in Task.objects.filter(workflow__pk=workflow_id).all():
            list_input = [tinput for tinput in task.data if tinput["id"] == "Alpha"][0]
            self.assertIsInstance(list_input["list"]["value"], list)

    def test_when_eu_region_selected_then_no_region_on_data(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        workflow_data = {
            "name": "uploader",
            "data": [
                ALPHA,
                BETA,
                GAMMA,
            ],
        }
        _ = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        workflow_id = Workflow.objects.get(name="uploader").id
        with open(self.file_path, encoding="ISO-8859-1") as f:
            data = {"file": f}
            response = self.client.post(
                f"/v1/orgs/{self.org_id}/workflows/{workflow_id}/upload?region=EU",
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

        task = Task.objects.filter(workflow__pk=workflow_id).first()
        self.assertIn(task.source.name, self.file_path)
        assert task.region is None

    @pytest.mark.bucket
    def test_when_non_eu_region_selected_then_confirmed_on_data(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        workflow_data = {
            "name": "uploader",
            "data": [
                ALPHA,
                BETA,
                GAMMA,
            ],
        }
        _ = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        workflow_id = Workflow.objects.get(name="uploader").id
        with open(self.file_path, encoding="ISO-8859-1") as f:
            data = {"file": f}
            response = self.client.post(
                f"/v1/orgs/{self.org_id}/workflows/{workflow_id}/upload?region=AU",
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

        task = Task.objects.filter(workflow__pk=workflow_id).first()
        self.assertIn(task.source.name, self.file_path)
        assert task.region == "AU"



import copy
import logging
import os

from rest_framework import status

from human_lambdas.user_handler.models import Notification, Organization, User
from human_lambdas.workflow_handler.models import Task, Workflow
from human_lambdas.workflow_handler.tests.constants import (
    ALPHA,
    BETA,
    GAMMA,
    INTERNAL_WORKER_REGISTRATION_DATA,
    REGISTRATION_DATA,
    SUPER_ADMIN_REGISTRATION_DATA,
    WORKFLOW_DATA_3,
)
from human_lambdas.workflow_handler.tests.util import HLTestCase

logger = logging.getLogger(__name__)

_CURRENT_DIR = os.path.dirname(__file__)


class TestTasks(HLTestCase):
    def setUserClient(self, email):
        response = self.client.post(
            "/v1/users/token", {"email": email, "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)

    def setUp(self):
        self.access_token_super_admin = self.register(
            SUPER_ADMIN_REGISTRATION_DATA, is_super_admin=True
        )["access"]
        self.access_token_internal_worker = self.register(
            INTERNAL_WORKER_REGISTRATION_DATA, is_internal_worker=True
        )["access"]
        self.file_path = os.path.join(_CURRENT_DIR, "data", "test.csv")

        response = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.user_id = response.data["id"]
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk

        # create user and add in same org
        org = Organization.objects.get(pk=self.org_id)
        notification = Notification()
        notification.save()
        user = User(name="foo", email="foojr@bar.com", notifications=notification)
        user.set_password("foowordbar")
        user.save()
        org.add_admin(user)

        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)

        _ = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            WORKFLOW_DATA_3,
            format="json",
        )
        second_workflow_data = {
            "name": "second workflow",
            "data": [
                ALPHA,
                BETA,
                GAMMA,
                {
                    "id": "bar",
                    "name": "bar",
                    "type": "multiple_selection",
                    "multiple_selection": {
                        "options": [
                            {"id": "foo2", "name": "foo2"},
                            {"id": "bar2", "name": "bar2"},
                        ],
                    },
                },
            ],
        }
        _ = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            second_workflow_data,
            format="json",
        )
        self.workflow_id = Workflow.objects.get(name="uploader").id
        self.second_workflow_id = Workflow.objects.get(name="second workflow").id
        with open(self.file_path) as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(
                    self.org_id, self.workflow_id
                ),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)
        with open(self.file_path) as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(
                    self.org_id, self.second_workflow_id
                ),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

    def test_get_task_list(self):
        response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}/tasks".format(self.org_id, self.workflow_id)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)
        self.assertEqual(3, len(response.data), response.content)

    def test_retrieve_task(self):
        workflow = Workflow.objects.get(id=self.workflow_id)
        tasks = Task.objects.filter(workflow=workflow).all()
        for task in tasks:
            response = self.client.get(
                "/v1/orgs/{0}/workflows/{1}/tasks/{2}".format(
                    self.org_id, self.workflow_id, task.id
                )
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)
            self.assertEqual("in_progress", response.data["status"])
            self.assertTrue(len(response.data["data"]))
            for itask in response.data["data"]:
                if itask[itask["type"]].get("read_only"):
                    self.assertTrue("value" in itask[itask["type"]])

    def test_complete_single_selection_task(self):
        workflow = Workflow.objects.get(id=self.workflow_id)
        tasks = Task.objects.filter(workflow=workflow).all()
        output_values = ["foo1", "bar1"]
        for output_value, task in zip(output_values, tasks):
            data = copy.deepcopy(task.data)

            _ = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/tasks/{2}/assign".format(
                    self.org_id, self.workflow_id, task.id
                ),
                data={"assigned_to": self.user_id},
                format="json",
            )

            for idata in data:
                if idata["id"] == "foo":
                    idata["single_selection"]["value"] = output_value
            data_payload = {"data": data, "assigned_to": self.user_id}

            response = self.client.patch(
                "/v1/orgs/{0}/workflows/{1}/tasks/{2}".format(
                    self.org_id, self.workflow_id, task.id
                ),
                data=data_payload,
                format="json",
            )

            self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)
            self.assertEqual("completed", response.data["status"])
            self.assertEqual(data, response.data["data"])
            self.assertEqual("completed", Task.objects.get(id=task.id).status)

            # test retrieving the same task
            response = self.client.get(
                "/v1/orgs/{0}/workflows/{1}/tasks/{2}".format(
                    self.org_id, self.workflow_id, task.id
                )
            )
            self.assertEqual("completed", response.data["status"])
            self.assertEqual(data, response.data["data"])

    def test_complete_multiple_selection_task(self):
        workflow = Workflow.objects.get(id=self.second_workflow_id)
        tasks = Task.objects.filter(workflow=workflow).all()
        output_values = [["foo1"], ["foo1", "bar1"]]
        for output_value, task in zip(output_values, tasks):
            data = copy.deepcopy(task.data)

            _ = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/tasks/{2}/assign".format(
                    self.org_id, self.second_workflow_id, task.id
                ),
                data={"assigned_to": self.user_id},
                format="json",
            )

            for idata in data:
                if idata["id"] == "bar":
                    idata["multiple_selection"]["value"] = output_value
            data_reponse = {"data": data}  # , "assigned_to": self.user_id}
            response = self.client.patch(
                "/v1/orgs/{0}/workflows/{1}/tasks/{2}".format(
                    self.org_id, self.second_workflow_id, task.id
                ),
                data=data_reponse,
                format="json",
            )

            self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)
            self.assertEqual("completed", response.data["status"])
            self.assertEqual(data, response.data["data"])
            self.assertEqual("completed", Task.objects.get(id=task.id).status)

    def test_next_task(self):
        response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}/tasks/next".format(
                self.org_id, self.workflow_id
            )
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)
        self.assertEqual("in_progress", response.data["status"], response.content)
        task = Task.objects.get(id=response.data["id"])
        self.assertEqual(task.status, "in_progress")

    def test_next_task_super_admin(self):
        self.client.credentials(
            HTTP_AUTHORIZATION="Bearer " + self.access_token_super_admin
        )
        response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}/tasks/next".format(
                self.org_id, self.workflow_id
            )
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)
        self.assertEqual("in_progress", response.data["status"], response.content)
        task = Task.objects.get(id=response.data["id"])
        self.assertEqual(task.status, "in_progress")

    def test_next_task_internal_worker(self):
        self.client.credentials(
            HTTP_AUTHORIZATION="Bearer " + self.access_token_internal_worker
        )
        response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}/tasks/next".format(
                self.org_id, self.workflow_id
            )
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)
        self.assertEqual("in_progress", response.data["status"], response.content)
        task = Task.objects.get(id=response.data["id"])
        self.assertEqual(task.status, "in_progress")

    def test_next_task_repeatedly(self):
        response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}/tasks/next".format(
                self.org_id, self.workflow_id
            )
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)
        self.assertEqual("in_progress", response.data["status"], response.content)

        task_id = response.data["id"]
        response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}/tasks/next".format(
                self.org_id, self.workflow_id
            )
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)
        self.assertEqual(response.data["id"], task_id)

        # get next task by client 2 should be different
        self.setUserClient("foojr@bar.com")
        response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}/tasks/next".format(
                self.org_id, self.workflow_id
            )
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)
        self.assertNotEqual(response.data["id"], task_id)

    def test_next_task_different_workflow(self):
        response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}/tasks/next".format(
                self.org_id, self.second_workflow_id
            )
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

    def test_next_task_with_slash(self):
        response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}/tasks/next/".format(
                self.org_id, self.workflow_id
            )
        )
        self.assertEqual(
            response.status_code, status.HTTP_404_NOT_FOUND, response.content
        )

    def test_workflow_list_task_count(self):
        response = self.client.get("/v1/orgs/{}/workflows".format(self.org_id))
        for data, expected in zip(response.data, [3, 3]):
            self.assertEqual(data["n_tasks"], expected)

    def test_workflow_retrieve_task_count(self):
        response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}".format(self.org_id, self.workflow_id)
        )
        self.assertEqual(response.data["n_tasks"], 3)

    def test_change_outputs_retrieve_task(self):
        workflow = Workflow.objects.get(id=self.workflow_id)
        workflow.data[0]["name"] = "new_name"
        workflow.save()
        tasks = Task.objects.filter(workflow=workflow).all()
        for task in tasks:
            response = self.client.get(
                "/v1/orgs/{0}/workflows/{1}/tasks/{2}".format(
                    self.org_id, self.workflow_id, task.id
                )
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)
            self.assertEqual(response.data["data"][0]["name"], "new_name")

    def test_next_task_change_data(self):
        workflow = Workflow.objects.get(id=self.workflow_id)
        workflow.data[0]["name"] = "new_name"
        workflow.save()
        response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}/tasks/next".format(
                self.org_id, self.workflow_id
            )
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)
        self.assertEqual(response.data["data"][0]["name"], "new_name")

    def test_get_pending_tasks(self):
        response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}/tasks/pending".format(
                self.org_id,
                self.workflow_id,
            )
        )
        self.assertEqual(response.data["count"], 3)
        self.assertEqual(response.data["next"], None)
        for idata in response.data["tasks"]:
            assert idata["data"] != None
            self.assertEqual(idata["status"], "new")

        response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}/tasks/next".format(
                self.org_id, self.workflow_id
            )
        )
        task_id = response.data["id"]
        task_data = response.data["data"]

        response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}/tasks/pending".format(
                self.org_id,
                self.workflow_id,
            )
        )
        self.assertEqual(response.data["count"], 3)
        self.assertEqual(
            len(
                [
                    idata
                    for idata in response.data["tasks"]
                    if idata["status"] == "in_progress"
                ]
            ),
            1,
        )
        for t_data in task_data:
            if t_data["id"] == "foo":
                t_data[t_data["type"]]["value"] = "foo2"
        data = {"data": task_data}
        response = self.client.patch(
            "/v1/orgs/{0}/workflows/{1}/tasks/{2}".format(
                self.org_id, self.workflow_id, task_id
            ),
            data=data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

        response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}/tasks/pending".format(
                self.org_id,
                self.workflow_id,
            )
        )
        self.assertEqual(response.data["count"], 2)
        self.assertEqual(response.data["tasks"][0]["n_comments"], 0)



from django.urls import path

from human_lambdas.workflow_handler.flush import FlushTasksView
from human_lambdas.workflow_handler.views import (
    ExternalWorkflowView,
    ListWorkflowView,
    RUWebhookView,
)

from .views import CreateTaskView, GetExternalCompletedTaskView
from .zapier import (
    GetZapierTaskInputs,
    GetZapierTaskSampleData,
    GetZapierWorkflows,
    ZapierAuthentication,
    ZapierCreateTask,
    ZapierHook,
)

urlpatterns = [
    path(
        "orgs/<int:org_id>/queues",
        ListWorkflowView.as_view(),
        name="create-task",
    ),
    path(
        "orgs/<int:org_id>/queues/<int:queue_id>",
        ExternalWorkflowView.as_view(),
        name="list-workflows",
    ),
    path(
        "orgs/<int:org_id>/queues/<int:queue_id>/webhook",
        RUWebhookView.as_view(),
        name="webhook",
    ),
    path(
        "orgs/<int:org_id>/queues/<int:queue_id>/tasks/create",
        CreateTaskView.as_view(),
        name="create-task",
    ),
    path(
        "orgs/<int:org_id>/queues/<int:queue_id>/tasks/completed",
        GetExternalCompletedTaskView.as_view(),
        name="completed-task",
    ),
    path(
        "orgs/<int:org_id>/queues/<int:queue_id>/flush",
        FlushTasksView.as_view(),
        name="flush-task",
    ),
    # Legacy workflows terminology
    path(
        "orgs/<int:org_id>/workflows",
        ListWorkflowView.as_view(),
        name="create-task",
    ),
    path(
        "orgs/<int:org_id>/workflows/<int:workflow_id>",
        ExternalWorkflowView.as_view(),
        name="list-workflows",
    ),
    path(
        "orgs/<int:org_id>/workflows/<int:workflow_id>/webhook",
        RUWebhookView.as_view(),
        name="webhook",
    ),
    path(
        "orgs/<int:org_id>/workflows/<int:workflow_id>/tasks/create",
        CreateTaskView.as_view(),
        name="create-task",
    ),
    path(
        "orgs/<int:org_id>/workflows/<int:workflow_id>/tasks/completed",
        GetExternalCompletedTaskView.as_view(),
        name="completed-task",
    ),
    path(
        "orgs/<int:org_id>/workflows/<int:workflow_id>/flush",
        FlushTasksView.as_view(),
        name="flush-task",
    ),
    # Zapier Endpoints
    path(
        "zapier/authentication",
        ZapierAuthentication.as_view(),
        name="zapier-authentication",
    ),
    path("zapier/get-workflows", GetZapierWorkflows.as_view(), name="zapier-workflows"),
    path(
        "zapier/get-task-inputs",
        GetZapierTaskInputs.as_view(),
        name="zapier-task-inputs",
    ),
    path(
        "zapier/<int:workflow_id>/create-task",
        ZapierCreateTask.as_view(),
        name="zapier-create-task",
    ),
    path(
        "zapier/hook",
        ZapierHook.as_view(),
        name="zapier-hook",
    ),
    path("zapier/sample-data", GetZapierTaskSampleData.as_view(), name="sample-data"),
]



import copy
import datetime
import logging
import os
from typing import Any, Dict

import cchardet
import requests
import sentry_sdk
from django.utils import timezone
from django.utils.timezone import make_aware
from rest_framework.pagination import LimitOffsetPagination
from rest_framework.request import Request
from rest_framework.response import Response

from .models import WebHook, Workflow, WorkflowNotification

logger = logging.getLogger(__name__)

TEMPLATE_ORG_ID = 40
STAFF_ORG_ID = 1000000000


def parse_dates(request):
    try:
        start_date = datetime.datetime.fromisoformat(
            request.query_params.get("start_date")
        )  # YYYY-MM-DD
    except:
        start_date = datetime.datetime(2019, 1, 1)
    try:
        end_date = datetime.datetime.fromisoformat(
            request.query_params.get("end_date")
        )  # YYYY-MM-DD
        end_date += datetime.timedelta(days=1, microseconds=-1)
    except:
        end_date = datetime.datetime.today()
    return make_aware(start_date), make_aware(end_date)


def find_and_fire_hook(event_name, instance, **kwargs):
    filters = {
        "event": event_name,
        "workflow": instance.workflow,
    }
    hooks = WebHook.objects.filter(**filters)
    for hook in hooks:
        hook.deliver_hook(instance)


def decode_csv(file_obj):
    for line in file_obj:
        yield unidecode(line)


def unidecode(text):
    try:
        return text.decode("utf-8")
    except UnicodeDecodeError:
        return text.decode(cchardet.detect(text)["encoding"])


def process_filter_value(filter_value):
    if filter_value == "true":
        return True
    if filter_value == "false":
        return False
    return filter_value


def process_query_params(query_params):
    filter_mapper = [
        ("workflow__pk", "queue_id"),
        ("assigned_to__pk", "worker_id"),
        ("source__pk", "source_id"),
        ("correct", "correct"),
    ]
    filters = {}
    for filter_name, param_name in filter_mapper:
        filter_value = query_params.get(param_name)
        if filter_value:
            filters[filter_name] = process_filter_value(filter_value)
    return filters


class TaskPagination(LimitOffsetPagination):
    """
    Extended pagination class for Tasks
    """

    default_limit = 100
    max_limit = 1000

    def get_paginated_response(self, data):
        return Response(
            {
                "next": self.get_next_link(),
                "previous": self.get_previous_link(),
                "count": self.count,
                "tasks": data,
            },
            status=200,
        )


def create_template(template_id, user, org):
    if template_id:
        workflow = Workflow.objects.filter(
            organization=TEMPLATE_ORG_ID, pk=template_id
        ).first()
        if workflow:
            Workflow(
                name=workflow.name,
                created_by=user,
                data=copy.deepcopy(workflow.data),
                organization=org,
            ).save()
            for org_user in org.user.all():
                wfnotification = WorkflowNotification(
                    workflow=workflow, notification=org_user.notifications, enabled=True
                )
                wfnotification.save()


def get_session_duration_seconds(task):
    if task.session_started_at:
        delta_since_start = timezone.now() - task.session_started_at
        return delta_since_start / timezone.timedelta(seconds=1)

    return 0


def is_force(query_params):
    force = query_params.get("force")
    if force:
        return force.lower() == "true"
    return False


def notify_staff_run_status(data: Dict[str, Any], request: Request):
    if not "SLACK_WEBHOOK_URL" in os.environ:
        logger.warn("not sending notification, slack not configured")
        return
    try:
        status = "running" if data["is_running"] else "paused"
        text = f" set {data['name']} to {status}"
        notify_slack(text, request)
    except Exception as ex:
        sentry_sdk.capture_exception(ex)


def notify_slack(text: str, request: Request):
    if not "SLACK_WEBHOOK_URL" in os.environ:
        logger.warn("not sending notification, slack not configured")
        return
    try:
        email = request.user.email
        path = request.stream.path
        slack_url = os.getenv("SLACK_WEBHOOK_URL")
        if slack_url is None:
            sentry_sdk.capture_exception("No slack url configured!")

        r = requests.post(slack_url, json={"text": f"{email} called {path}: {text}"})
        if r.status_code != 200:
            return sentry_sdk.capture_message(
                f"Slack notification failed with status {r.status_code}"
            )
    except Exception as ex:
        sentry_sdk.capture_exception(ex)



import logging

from django.conf import settings
from sendgrid import SendGridAPIClient
from sendgrid.helpers.mail import Asm, Email, GroupId
from sendgrid.helpers.mail import Mail as SGMail
from sendgrid.helpers.mail import Personalization

logger = logging.getLogger(__name__)


def get_personalization(email, template_data):
    personalization = Personalization()
    personalization.add_to(email)
    personalization.dynamic_template_data = template_data
    return personalization


class SendGrid(SGMail):
    api_key = None
    client = None
    default_from = None

    def __init__(self, app=None, **opts):
        self.api_key = settings.SENDGRID_API_KEY
        self.default_from = ("noreply@humanlambdas.com", "Human Lambdas")
        self.client = SendGridAPIClient(self.api_key).client
        self.reset_data()

    def reset_data(self):
        super(SGMail, self).__init__()
        self.from_email = None
        self.subject = None
        self._personalizations = None
        self._contents = None
        self._attachments = None
        self._template_id = None
        self._sections = None
        self._headers = None
        self._categories = None
        self._custom_args = None
        self._send_at = None
        self._batch_id = None
        self._asm = None
        self._ip_pool_name = None
        self._mail_settings = None
        self._tracking_settings = None
        self._reply_to = None

    def send_email(
        self,
        to_email,
        template_id,
        template_data,
        group_id,
        from_email=None,
        *args,
        **kwargs,
    ):  # noqa
        if settings.DEBUG or not to_email:
            return logger.info(
                f"Will not send any email but here is the template data: {template_data}"
            )
        if not any([from_email, self.default_from]):
            raise ValueError("Missing from email and no default.")
        self.template_id = template_id
        self.from_email = self.default_from
        if type(to_email) is list:
            if isinstance(template_data, list):
                for email, itemplate in zip(
                    self._extract_emails(to_email), template_data
                ):
                    self.add_personalization(get_personalization(email, itemplate))
            else:
                for email in self._extract_emails(to_email):
                    self.add_personalization(get_personalization(email, template_data))
        elif type(to_email) is Email:
            self.add_personalization(get_personalization(to_email, template_data))
        elif type(to_email) is str:
            self.add_personalization(
                get_personalization(Email(to_email), template_data)
            )

        self.asm = Asm(GroupId(group_id))
        request_body = self.get()
        logger.info("Sending email with request body: %s", request_body)
        sent = self.client.mail.send.post(request_body=request_body)
        self.reset_data()
        return sent

    def _extract_emails(self, emails):
        if type(emails[0]) is Email:
            for email in emails:
                yield email
        elif type(emails[0]) is dict:
            for email in emails:
                yield Email(email["email"])
        else:
            for email in emails:
                yield Email(email)



import copy
import logging
import os

from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization
from human_lambdas.workflow_handler.models import Task, Workflow
from human_lambdas.workflow_handler.tests.constants import (
    REGISTRATION_DATA,
    WORKFLOW_DATA_3,
)

from . import DATA_PATH

logger = logging.getLogger(__name__)


class TestTaskCount(APITestCase):
    def setUp(self):
        self.file_path = os.path.join(DATA_PATH, "test.csv")
        self.total_rows = 3

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get(
            "/v1/users/api-token",
        )
        self.token = response.data["token"]

        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            WORKFLOW_DATA_3,
            format="json",
        )
        self.workflow_id = response.data["id"]
        with open(self.file_path) as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(
                    self.org_id, self.workflow_id
                ),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

    def test_total_count(self):
        workflow = Workflow.objects.get(pk=self.workflow_id)
        self.assertEqual(workflow.n_tasks, self.total_rows)

    def test_assigning_task(self):
        n_tasks = self.total_rows
        for i in range(self.total_rows):
            response = self.client.get(
                "/v1/orgs/{}/workflows/{}/tasks/next".format(
                    self.org_id, self.workflow_id
                )
            )
            workflow = Workflow.objects.get(pk=self.workflow_id)
            self.assertEqual(workflow.n_tasks, n_tasks)
            for idata in response.data["data"]:
                if idata["id"] == "foo":
                    idata[idata["type"]]["value"] = "bajs"
            data = {"data": response.data["data"]}
            _ = self.client.patch(
                "/v1/orgs/{}/workflows/{}/tasks/{}".format(
                    self.org_id, self.workflow_id, response.data["id"]
                ),
                data=data,
                format="json",
            )
            n_tasks -= 1
            workflow = Workflow.objects.get(pk=self.workflow_id)
            self.assertEqual(workflow.n_tasks, n_tasks)

    def test_task_creation(self):
        self.client.credentials(HTTP_AUTHORIZATION="Token " + self.token)
        n_tasks = self.total_rows
        task_data = {
            "data": {"Alpha": "data1", "Beta": "data2", "Gamma": "data3"},
        }
        for i in range(5):
            response = self.client.post(
                "/orgs/{}/workflows/{}/tasks/create".format(
                    self.org_id, self.workflow_id
                ),
                task_data,
                format="json",
            )
            n_tasks += 1
            workflow = Workflow.objects.get(pk=self.workflow_id)
            self.assertEqual(workflow.n_tasks, n_tasks)

        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get(
            "/v1/orgs/{}/workflows/{}/tasks/{}".format(
                self.org_id, self.workflow_id, response.data["id"]
            )
        )

        workflow = Workflow.objects.get(pk=self.workflow_id)
        self.assertEqual(workflow.n_tasks, n_tasks)
        data = copy.deepcopy(response.data["data"])
        for idata in data:
            if idata["id"] == "foo":
                idata["single_selection"]["value"] = "bar2"
        response_data = {"data": data}
        _ = self.client.patch(
            "/v1/orgs/{}/workflows/{}/tasks/{}".format(
                self.org_id, self.workflow_id, response.data["id"]
            ),
            data=response_data,
            format="json",
        )
        n_tasks -= 1
        workflow = Workflow.objects.get(pk=self.workflow_id)
        self.assertEqual(workflow.n_tasks, n_tasks)

    def test_unassigning_task(self):
        n_tasks = self.total_rows
        response = self.client.get(
            "/v1/orgs/{}/workflows/{}/tasks/next".format(self.org_id, self.workflow_id)
        )
        workflow = Workflow.objects.get(pk=self.workflow_id)
        self.assertEqual(workflow.n_tasks, n_tasks)
        _ = self.client.post(
            "/v1/orgs/{}/workflows/{}/tasks/{}/assign".format(
                self.org_id,
                self.workflow_id,
                response.data["id"],
            ),
            data={"assigned_to": None},
            format="json",
        )
        workflow = Workflow.objects.get(pk=self.workflow_id)
        self.assertEqual(workflow.n_tasks, n_tasks)
        task = Task.objects.get(pk=response.data["id"])
        self.assertEqual(task.status, "open")
        self.assertEqual(task.assigned_at, None)

    def test_active_users(self):
        n_tasks = self.total_rows
        response = self.client.get(
            "/v1/orgs/{}/workflows/{}/tasks/next".format(self.org_id, self.workflow_id)
        )

        workflow = Workflow.objects.get(pk=self.workflow_id)
        self.assertEqual(workflow.n_tasks, n_tasks)
        response = self.client.get("/v1/orgs/{}/workflows".format(self.org_id))
        self.assertEqual(response.data[0]["active_users"], ["foo"])



import analytics
from django.apps import AppConfig
from django.conf import settings

from human_lambdas.hl_rest_api.email_client import SendGrid


class UserHandlerConfig(AppConfig):
    name = "human_lambdas.user_handler"
    emailclient = SendGrid()

    def ready(self):
        analytics.write_key = settings.SEGMENT_KEY



from typing import Any, Dict, List
from unittest.mock import patch

import pytest
from django.http.request import HttpRequest
from django.http.response import HttpResponse

from human_lambdas.workflow_handler.latency import (
    LatencyMiddleware,
    QueryLogger,
)


def test_when_query_logger_invoked_then_okay():
    ql = QueryLogger()

    def mock_execute(*args: List[Any], **kwargs: Dict[Any, Any]):
        return 42

    assert ql(mock_execute, "", {}, False, {}) == 42


def test_when_query_logger_exception_then_raised():
    ql = QueryLogger()

    def mock_execute(*args: List[Any], **kwargs: Dict[Any, Any]):
        raise ValueError()

    with pytest.raises(ValueError):
        ql(mock_execute, "", {}, False, {})


def test_when_latency_middleware_next_then_logs():
    with patch("human_lambdas.workflow_handler.latency.logger.info") as info:
        request = HttpRequest()
        request.path = "asdf/next"

        lm = LatencyMiddleware()
        lm.process_request(request)

        lm.process_response(request, HttpResponse())
        assert info.call_count > 0


def test_when_latency_middleware_not_next_then_no_log():
    with patch("human_lambdas.workflow_handler.latency.logger.info") as info:
        request = HttpRequest()
        request.path = "asdf"

        lm = LatencyMiddleware()
        lm.process_request(request)

        lm.process_response(request, HttpResponse())
        assert info.call_count == 0



# Generated by Django 2.2.5 on 2019-12-27 13:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("user_handler", "0003_auto_20191119_1829"),
    ]

    operations = [
        migrations.AlterField(
            model_name="organization",
            name="name",
            field=models.CharField(max_length=128, unique=True),
        ),
    ]



import logging
import time
from contextlib import contextmanager
from typing import Any, Callable, Collection, Dict, Optional

from django.http.request import HttpRequest
from django.http.response import HttpResponse
from django.utils.deprecation import MiddlewareMixin

logger = logging.getLogger(__name__)


class QueryLogger:
    """
    https://docs.djangoproject.com/en/3.1/topics/db/instrumentation/
    """

    def __call__(
        self,
        execute: Callable[..., Any],
        sql: str,
        params: Collection[Any],
        many: bool,
        context: Dict[Any, Any],
    ):
        current_query: Dict[str, Any] = {"sql": sql, "params": params, "many": many}
        start = time.monotonic()
        try:
            result = execute(sql, params, many, context)
        except Exception as e:
            current_query["status"] = "error"
            current_query["exception"] = e
            raise
        else:
            current_query["status"] = "ok"
            duration = time.monotonic() - start
            current_query["duration"] = round(duration * 1000, 3)
            return result


class LatencyMiddleware(MiddlewareMixin):
    def process_request(self, request: HttpRequest):
        if request.path.endswith("next"):
            request.start_time = time.monotonic()

    def process_response(self, request: HttpRequest, response: HttpResponse):
        if request.path.endswith("next"):
            end = time.monotonic()
            total: float = end - request.start_time
            logger.info(f"/next django latency: {latency_str(total)}")

        return response


def latency_str(latency: float):
    return f"{round(latency * 1000, 3)} ms"


@contextmanager
def timer(label: Optional[str]):
    ss = time.monotonic()
    try:
        yield
    finally:
        ee = time.monotonic()
        duration = ee - ss
        logger.info(f"{label}: {round(duration * 1000)}ms")



# Generated by Django 2.2.13 on 2020-10-21 15:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0027_auto_20200918_1421"),
    ]

    operations = [
        migrations.AlterField(
            model_name="task",
            name="status",
            field=models.CharField(default="new", max_length=128),
        ),
    ]



import os
from datetime import timedelta
from unittest.mock import patch

from django.utils import timezone
from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization
from human_lambdas.workflow_handler.models import Task, User, Workflow
from human_lambdas.workflow_handler.tests import DATA_PATH
from human_lambdas.workflow_handler.tests.constants import (
    ALPHA,
    BETA,
    GAMMA,
    REGISTRATION_DATA,
    WORKFLOW_DATA_3,
)


class TestHandlingTime(APITestCase):
    def setUp(self):
        self.file_path = os.path.join(DATA_PATH, "test.csv")
        self.total_rows = 3

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get(
            "/v1/users/api-token",
        )
        self.token = response.data["token"]

        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            WORKFLOW_DATA_3,
            format="json",
        )
        self.workflow_id = response.data["id"]
        with open(self.file_path) as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(
                    self.org_id, self.workflow_id
                ),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

        workflow_data = {
            "name": "uploader2",
            "data": [
                ALPHA,
                BETA,
                GAMMA,
                {
                    "id": "foo",
                    "name": "foo",
                    "type": "single_selection",
                    "single_selection": {
                        "options": [
                            {"id": "foo2", "name": "foo2"},
                            {"id": "bar2", "name": "bar2"},
                        ],
                    },
                },
            ],
        }
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.workflow_id2 = response.data["id"]
        with open(self.file_path) as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(
                    self.org_id, self.workflow_id2
                ),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

        wf1 = Workflow.objects.get(pk=self.workflow_id)
        self.task = Task.objects.filter(workflow=wf1).first()
        self.task.status = "completed"

        assigned_at = timezone.now() - timezone.timedelta(days=7)
        created_at = assigned_at - timezone.timedelta(days=1)
        self.task.assigned_at = assigned_at
        self.task.assigned_to = User.objects.get(email="foo@bar.com")
        self.task.created_at = created_at
        self.task.save()

    def _start_session(self):
        response = self.client.get(
            f"/v1/orgs/{self.org_id}/workflows/{self.workflow_id}/tasks/{self.task.id}/activity"
        )
        assert response.status_code == 200

    def _save_task(self):
        data_payload = {"data": self.task.data}

        response = self.client.patch(
            f"/v1/orgs/{self.org_id}/workflows/{self.workflow_id}/tasks/{self.task.id}/save?force=true",
            data=data_payload,
            format="json",
        )
        assert response.status_code == 200

    def _submit_task(self):
        response = self.client.patch(
            f"/v1/orgs/{self.org_id}/workflows/{self.workflow_id}/tasks/{self.task.id}?force=true",
            data={},
            format="json",
        )
        assert response.status_code == 200

    def _post_comment(self):
        response = self.client.post(
            "/v1/orgs/{0}/workflows/{1}/tasks/{2}/activity".format(
                self.org_id,
                self.workflow_id,
                self.task.pk,
            ),
            data={"action": "comment", "comment": "hello world"},
            format="json",
        )
        self.activity_id = response.data["id"]
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def _delete_comment(self):
        self.assertIsNotNone(self.activity_id)
        response = self.client.delete(
            "/v1/orgs/{0}/workflows/{1}/tasks/{2}/activity/{3}".format(
                self.org_id, self.workflow_id, self.task.pk, self.activity_id
            ),
        )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_when_resumed_task_submits_then_ht_is_sum_of_sessions(self):
        # Arrange
        data = {
            "range": "monthly",
            "type": ["aht"],
        }

        first_ht = 2
        downtime = 8
        second_ht = 6
        total_ht = first_ht + second_ht
        start_time = timezone.now()

        with patch("django.utils.timezone.now") as now:
            # Act
            now.return_value = start_time
            self._start_session()

            now.return_value += timedelta(seconds=first_ht)
            self._save_task()

            now.return_value += timedelta(seconds=downtime)
            self._start_session()

            now.return_value += timedelta(seconds=second_ht)

            self._submit_task()

            response = self.client.get(
                "/v1/orgs/{}/metrics".format(self.org_id),
                data,
                format="json",
            )

            # Assert
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.data[-1]["aht"], total_ht)

    def test_when_comment_posted_then_ht_includes_session(self):
        # Arrange
        data = {
            "range": "monthly",
            "type": ["aht"],
        }

        first_ht = 2
        downtime = 8
        second_ht = 6
        total_ht = first_ht + second_ht
        start_time = timezone.now()

        with patch("django.utils.timezone.now") as now:
            # Act
            now.return_value = start_time
            self._start_session()

            now.return_value += timedelta(seconds=first_ht)
            self._post_comment()

            now.return_value += timedelta(seconds=downtime)
            self._start_session()

            now.return_value += timedelta(seconds=second_ht)

            self._submit_task()

            response = self.client.get(
                "/v1/orgs/{}/metrics".format(self.org_id),
                data,
                format="json",
            )

            # Assert
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.data[-1]["aht"], total_ht)

    def test_when_comment_deleted_then_ht_includes_session(self):
        # Arrange
        data = {
            "range": "monthly",
            "type": ["aht"],
        }

        first_ht = 2
        downtime = 8
        second_ht = 6
        third_ht = 3
        total_ht = first_ht + second_ht + third_ht
        start_time = timezone.now()

        with patch("django.utils.timezone.now") as now:
            # Act
            now.return_value = start_time
            self._start_session()

            now.return_value += timedelta(seconds=first_ht)
            self._post_comment()

            now.return_value += timedelta(seconds=downtime)
            self._start_session()

            now.return_value += timedelta(seconds=second_ht)
            self._delete_comment()

            now.return_value += timedelta(seconds=downtime)
            self._start_session()

            now.return_value += timedelta(seconds=third_ht)
            self._submit_task()

            response = self.client.get(
                "/v1/orgs/{}/metrics".format(self.org_id),
                data,
                format="json",
            )

            # Assert
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.data[-1]["aht"], total_ht)

    def test_when_page_refreshes_without_session_end_then_ht_not_counted(self):
        # Arrange
        data = {
            "range": "monthly",
            "type": ["aht"],
        }

        first_ht = 2
        downtime = 8
        second_ht = 6
        total_ht = second_ht
        start_time = timezone.now()

        with patch("django.utils.timezone.now") as now:
            # Act
            now.return_value = start_time

            self._start_session()
            now.return_value += timedelta(seconds=first_ht)
            # no session end

            now.return_value += timedelta(seconds=downtime)

            self._start_session()
            now.return_value += timedelta(seconds=second_ht)
            self._submit_task()

            response = self.client.get(
                "/v1/orgs/{}/metrics".format(self.org_id),
                data,
                format="json",
            )

            # Assert
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.data[-1]["aht"], total_ht)



from typing import Any, Dict, Optional

from django.core import management
from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization, User
from human_lambdas.workflow_handler.utils import STAFF_ORG_ID


class HLTestCase(APITestCase):
    org_id: Optional[str]

    def register(
        self,
        registration_data: Dict[str, str],
        is_super_admin: bool = False,
        is_internal_worker: bool = False,
    ) -> Dict[Any, Any]:
        reg_response = self.client.post("/v1/users/register", registration_data)
        token_response = self.client.post(
            "/v1/users/token",
            {
                "email": registration_data["email"],
                "password": registration_data["password"],
            },
        )
        if is_super_admin:
            management.call_command("createsuperadmin", registration_data["email"])
        elif is_internal_worker:
            templates_org, _created = Organization.objects.get_or_create(
                id=STAFF_ORG_ID, name="staff"
            )
            user = User.objects.get(email=registration_data["email"])
            templates_org.user.add(user)

        assert token_response.status_code == 200
        return {**token_response.data, **reg_response.data}

    def create_workflow(
        self, data: Dict[str, Any], org_id: Optional[str] = None
    ) -> Dict[str, Any]:
        response = self.client.post(
            f"/v1/orgs/{org_id or self.org_id}/workflows/create",
            data,
            format="json",
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        return response.data

    def list_workflows(self, org_id: Optional[str] = None) -> Dict[str, Any]:
        response = self.client.get(
            f"/v1/orgs/{org_id or self.org_id}/workflows",
            format="json",
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        return response.data

    def set_user(self, access_token: str, org_id: Optional[str] = None):
        self.org_id = org_id
        self.client.credentials(HTTP_AUTHORIZATION=f"Bearer {access_token}")

    def get_completed_tasks(self) -> Dict[str, Any]:
        response = self.client.get(
            "/v1/orgs/{}/workflows/tasks/completed".format(
                self.org_id,
            )
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        return response.data



# Generated by Django 2.2.5 on 2019-11-18 17:46

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="User",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("password", models.CharField(max_length=128, verbose_name="password")),
                (
                    "last_login",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="last login"
                    ),
                ),
                ("name", models.CharField(max_length=128)),
                ("email", models.EmailField(max_length=254)),
                ("is_admin", models.BooleanField(default=False)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Organization",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=128)),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="user_handler.User",
                    ),
                ),
            ],
        ),
    ]



import logging

from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import User

logger = logging.getLogger(__name__)


class TestEmailCase(APITestCase):
    def setUp(self):
        self.user_name1 = "foo"
        self.user_email1 = "foo@bar.com"
        self.user_name2 = "Foo"
        self.user_email2 = "Foo@bar.com"
        self.organization_name = "fooinc"
        self.user_password = "foowordbar"

    def test_user_model(self):
        user = User(name=self.user_name2, email=self.user_email2)
        user.set_password(self.user_password)
        user.save()
        uid = user.pk
        user = User.objects.get(pk=uid)
        self.assertEqual(user.email, self.user_email1)
        # user = User(name=self.user_name1, email=self.user_email1)

    def test_registration(self):
        response = self.client.post(
            "/v1/users/register",
            {
                "email": self.user_email2,
                "password": self.user_password,
                "name": self.user_name2,
                "organization": self.organization_name,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.data["email"], self.user_email1)

    def test_login1(self):
        response = self.client.post(
            "/v1/users/register",
            {
                "email": self.user_email2,
                "password": self.user_password,
                "name": self.user_name2,
                "organization": self.organization_name,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        response = self.client.post(
            "/v1/users/token",
            {"email": self.user_email1, "password": self.user_password},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response)
        response = self.client.post(
            "/v1/users/token",
            {"email": self.user_email2.upper(), "password": self.user_password},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response)

    def test_login2(self):
        response = self.client.post(
            "/v1/users/register",
            {
                "email": self.user_email1,
                "password": self.user_password,
                "name": self.user_name1,
                "organization": self.organization_name,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        response = self.client.post(
            "/v1/users/token",
            {"email": self.user_email2, "password": self.user_password},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response)
        response = self.client.post(
            "/v1/users/token",
            {"email": self.user_email2.upper(), "password": self.user_password},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response)



import logging

from rest_framework import serializers

from human_lambdas.hl_rest_api import analytics
from human_lambdas.workflow_handler.utils import create_template

from .models import Notification, Organization, User

logger = logging.getLogger(__name__)


# Serializers define the API representation.
class UserSerializer(serializers.ModelSerializer):
    organization = serializers.CharField(
        max_length=128,
        allow_blank=True,
        write_only=True,
        allow_null=True,
        required=False,
    )
    password = serializers.CharField(min_length=8, write_only=True, required=False)
    template_id = serializers.IntegerField(write_only=True, required=False)

    class Meta:
        model = User
        fields = [
            "name",
            "password",
            "email",
            "organization",
            "current_organization_id",
            "id",
            "template_id",
        ]
        extra_kwargs = {
            "password": {"write_only": True},
            "template_id": {"write_only": True},
            "id": {"read_only": True},
        }

    def create(self, validated_data):
        name = validated_data["name"]
        password = validated_data.get("password")
        email = validated_data["email"].lower()
        notification = Notification()
        notification.save()
        user_obj = User(name=name, email=email, notifications=notification)
        if password:
            user_obj.set_password(password)
        user_obj.save()
        organization_name = validated_data.get("organization")
        if not organization_name:
            organization_name = f"{name}'s organization"
        org = Organization(name=organization_name)
        org.save()
        org.add_admin(user_obj)
        user_obj.current_organization_id = org.pk
        user_obj.save()
        analytics.signup_events(user_obj, org)
        create_template(validated_data.get("template_id"), user_obj, org)
        return user_obj

    def update(self, instance, validated_data):
        instance.name = validated_data.get("name", instance.name)
        instance.email = validated_data.get("email", instance.email)
        instance.current_organization_id = validated_data.get(
            "current_organization_id", instance.current_organization_id
        )
        password = validated_data.get("password")
        if password:
            instance.set_password(password)
        instance.save()
        return instance


class APITokenUserSerializer(serializers.Serializer):
    def validate(self, attrs):
        request = self.context.get("request")
        if not request.user:
            msg = "Unable to log in with user."
            raise serializers.ValidationError(msg, code="authorization")
        attrs["user"] = request.user
        return attrs



import logging

from django.shortcuts import get_object_or_404
from rest_framework.generics import (
    CreateAPIView,
    RetrieveUpdateAPIView,
    RetrieveUpdateDestroyAPIView,
)
from rest_framework.permissions import AllowAny, IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from human_lambdas.external.models import Token
from human_lambdas.hl_rest_api import analytics

from .models import Organization, User
from .permissions import IsOrgAdmin
from .serializers import APITokenUserSerializer, UserSerializer

logger = logging.getLogger(__name__)


class RegistrationView(CreateAPIView):
    permission_classes = (AllowAny,)
    queryset = User.objects.all()
    serializer_class = UserSerializer

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        serializer.data["status_code"] = 201
        return Response(serializer.data, status=201, headers=headers)


class HelloView(APIView):
    permission_classes = (IsAuthenticated,)

    def get(self, request):
        content = {"message": "Hello, Worldzzzzz!"}
        return Response(content)


class RetrieveUpdateUserView(RetrieveUpdateAPIView):
    permission_classes = (IsAuthenticated,)
    queryset = User.objects.all()
    serializer_class = UserSerializer

    def get_queryset(self):
        user = self.request.user
        return User.objects.filter(pk=user.pk)

    def get_object(self):
        obj = get_object_or_404(self.get_queryset(), id=self.kwargs["pk"])
        return obj

    def retrieve(self, request, *args, **kwargs):
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        is_admin = (
            Organization.objects.filter(pk=instance.current_organization_id)
            .filter(admin=instance)
            .exists()
        )
        data = serializer.data
        data["is_admin"] = is_admin
        data["status_code"] = 200
        return Response(data, status=200)

    def update(self, request, *args, **kwargs):
        partial = kwargs.pop("partial", False)
        instance = self.get_object()
        # change password work
        if request.data.get("currentPassword") or request.data.get("password"):
            if "currentPassword" not in request.data or "password" not in request.data:
                error_message = (
                    "Both the current password and the new password must be provided"
                )
                return Response(
                    {"status_code": 400, "errors": [{"message": error_message}]},
                    status=400,
                )
            if not instance.check_password(request.data["currentPassword"]):
                error_message = (
                    "The current password provided does not match the user password"
                )
                return Response(
                    {"status_code": 400, "errors": [{"message": error_message}]},
                    status=400,
                )
        # throw any errors otherwise let through
        serializer = self.get_serializer(instance, data=request.data, partial=partial)
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)
        serializer.data["status_code"] = 200
        return Response(serializer.data, status=200)


class RetrieveUpdateRemoveUserOrgView(RetrieveUpdateDestroyAPIView):
    permission_classes = (IsAuthenticated, IsOrgAdmin)
    queryset = User.objects.all()
    serializer_class = UserSerializer

    def get_queryset(self):
        organization = (
            Organization.objects.filter(pk=self.kwargs["org_id"])
            .filter(admin=self.request.user)
            .first()
        )
        return User.objects.filter(organization=organization)

    def get_object(self):
        obj = get_object_or_404(self.get_queryset(), id=self.kwargs["user_id"])
        return obj

    def update(self, request, *args, **kwargs):
        user = self.get_object()
        if request.user == user:
            return Response(
                {
                    "status_code": 403,
                    "errors": [{"message": "You cannot change your own role"}],
                },
                status=403,
            )
        else:
            org = Organization.objects.get(pk=kwargs["org_id"])
            try:
                if request.data["admin"]:
                    org.admin.add(user)
                    analytics.track(
                        self.request.user.pk, "Change Role", {"role": "admin"}
                    )
                    analytics.identify(self.request.user.pk, {"is_admin": True})
                    return Response(
                        {
                            "status_code": 200,
                            "message": "Users role was changed to admin",
                        },
                        status=200,
                    )
                else:
                    org.admin.remove(user)
                    analytics.track(
                        self.request.user.pk, "Change Role", {"role": "worker"}
                    )
                    analytics.identify(self.request.user.pk, {"is_admin": False})
                    return Response(
                        {
                            "status_code": 200,
                            "message": "Users role was changed to worker",
                        },
                        status=200,
                    )
            except KeyError:
                return Response(
                    {
                        "status_code": 400,
                        "errors": [
                            {"message": "payload should have the boolean field 'admin'"}
                        ],
                    },
                    status=400,
                )

    def destroy(self, request, *args, **kwargs):
        user = self.get_object()
        if request.user == user:
            return Response(
                {
                    "status_code": 403,
                    "errors": [{"message": "You cannot delete yourself"}],
                },
                status=403,
            )
        else:
            all_orgs_member = Organization.objects.filter(user=user)
            if all_orgs_member.count() == 1:
                user.delete()
                analytics.track(
                    self.request.user.pk, "Removed from Org", {"orgs_left": False}
                )
                return Response(status=204)
            else:
                org = all_orgs_member.get(pk=kwargs["org_id"])
                org.user.remove(user)
                org.admin.remove(user)
                analytics.track(
                    self.request.user.pk, "Removed from Org", {"orgs_left": True}
                )
                return Response(status=204)


class APIAuthToken(APIView):
    serializer_class = APITokenUserSerializer
    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        serializer = self.serializer_class(
            data=request.data, context={"request": request}
        )
        serializer.is_valid(raise_exception=True)
        user = serializer.validated_data["user"]
        if user.current_organization_id:
            token, created = Token.objects.get_or_create(
                user=user, organization=user.get_current_organization()
            )
            return Response(
                {
                    "status_code": 200,
                    "token": token.key,
                    "user_id": user.pk,
                    "email": user.email,
                    "organization_id": user.current_organization_id,
                },
                status=200,
            )
        else:
            return Response(
                {
                    "status_code": 400,
                    "errors": [
                        {"message": "user does not have a current organization set"}
                    ],
                },
                status=400,
            )



# Generated by Django 2.2.10 on 2020-06-11 15:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0016_auto_20200504_2334"),
    ]

    operations = [
        migrations.AlterField(
            model_name="workflow",
            name="name",
            field=models.CharField(max_length=128),
        ),
    ]



from rest_framework.test import APITestCase

from human_lambdas.hl_rest_api.utils import (
    generate_unique_token,
    is_invalid_email,
)


class TestUtils(APITestCase):
    def test_is_email_valid(self):
        self.assertTrue(is_invalid_email("test@test"))
        self.assertTrue(is_invalid_email("testgmail.com"))
        self.assertTrue(is_invalid_email("test@gmail,me"))
        self.assertFalse(is_invalid_email("with+plus@gmail.com"))
        self.assertTrue(is_invalid_email("A@b@c@gmail.com"))
        self.assertFalse(is_invalid_email("mailhost!username@gmail.com"))
        self.assertFalse(is_invalid_email("_somename@gmail.com"))
        self.assertFalse(is_invalid_email("$A12345@gmail.com"))
        self.assertFalse(is_invalid_email("user.nametagsorting@gmail.com"))
        self.assertFalse(is_invalid_email("valid@valid.com"))
        self.assertFalse(is_invalid_email("vlad@founders.network"))
        self.assertTrue(is_invalid_email('""Fred Bloggs"@example.com'))

    def test_generate_unique_token(self):
        self.assertNotEqual(
            generate_unique_token("one", "two", "three"),
            generate_unique_token("one", "two", "three"),
        )



from subprocess import check_output

from django.core.management.base import BaseCommand

from human_lambdas.user_handler.models import Organization, User
from human_lambdas.workflow_handler.models import Workflow
from human_lambdas.workflow_handler.tests.constants import WORKFLOW_DATA
from human_lambdas.workflow_handler.utils import TEMPLATE_ORG_ID


class Command(BaseCommand):
    help = "NOT FOR PROD: Sets up template org in the DB"

    def handle(self, *args, **options):

        check_output("dev_tools/get-templates > templates.json", shell=True)

        templates_org = Organization(id=TEMPLATE_ORG_ID, name="templates")
        templates_org.save()
        admin_id = 999

        try:
            user = User.objects.get(id=admin_id)
        except User.DoesNotExist:
            user = User(id=admin_id, name="template_admin", email="templates@_.com")
            user.save()
            templates_org.user.add(user)

        Workflow(
            data=WORKFLOW_DATA,
            name="template1",
            created_by=user,
            organization=templates_org,
        ).save()

        self.stdout.write(self.style.SUCCESS("Done"))



# Generated by Django 2.2.13 on 2020-10-29 14:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0028_auto_20201021_1505"),
    ]

    operations = [
        migrations.AlterField(
            model_name="source",
            name="name",
            field=models.CharField(max_length=512),
        ),
    ]



import logging

from django.conf import settings
from django.utils import timezone
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework.views import APIView

from human_lambdas.hl_rest_api.utils import (
    generate_unique_token,
    is_invalid_email,
)

from .apps import UserHandlerConfig
from .models import ForgottenPassword, User

logger = logging.getLogger(__name__)


class SendForgottenPasswordView(APIView):
    permission_classes = (AllowAny,)

    def post(self, request):
        if is_invalid_email(request.data["email"]):
            return Response(
                {
                    "status_code": 400,
                    "errors": [{"message": "This email is not valid"}],
                },
                status=400,
            )

        if not User.objects.filter(email=request.data["email"]).exists():
            return Response(
                {"status_code": 404, "errors": [{"message": "Email does not exist"}]},
                status=200,
            )
        token = generate_unique_token(request.data["email"])
        ForgottenPassword(
            email=request.data["email"],
            token=token,
            expires_at=timezone.now() + timezone.timedelta(15),
        ).save()
        password_link = "{0}forgot/{1}".format(settings.FRONT_END_BASE_URL, token)
        template_data = {"url": password_link}

        logger.info(
            f"password reset link for {request.data['email']} is: {password_link}"
        )

        if settings.ACCOUNT_ASM_GROUPID:
            UserHandlerConfig.emailclient.send_email(
                to_email=request.data["email"],
                template_id=settings.FORGOTTEN_PASSWORD_TEMPLATE,
                template_data=template_data,
                group_id=int(settings.ACCOUNT_ASM_GROUPID),
            )
        return Response(
            {"status_code": 200, "message": "Password Recovery Link Created"},
            status=200,
        )


class ForgottenPasswordView(APIView):
    permission_classes = (AllowAny,)

    def get(self, request, *args, **kwargs):
        forgotten_password = ForgottenPassword.objects.filter(
            token=self.kwargs["token"]
        )
        if forgotten_password.exists():
            return Response(
                {"status_code": 200},
                status=200,
            )
        else:
            return Response(
                {
                    "status_code": 404,
                    "errors": [
                        {"message": "this forgotten password token is not valid"}
                    ],
                },
                status=404,
            )

    def post(self, request, *args, **kwargs):
        forgotten_password = ForgottenPassword.objects.filter(token=kwargs["token"])
        if not forgotten_password.exists():
            return Response(
                {"status_code": 400, "errors": [{"message": "this token is invalid"}]},
                status=400,
            )
        if "password" not in self.request.data:
            return Response(
                {
                    "status_code": 400,
                    "errors": [{"message": "no password has been attached"}],
                },
                status=400,
            )
        if len(self.request.data["password"]) < 8:
            return Response(
                {
                    "status_code": 400,
                    "errors": [{"message": "passwords must be at least 8 characters"}],
                },
                status=400,
            )
        if forgotten_password.first().expires_at < timezone.now():
            return Response(
                {
                    "status_code": 400,
                    "errors": [{"message": "this token has expired!"}],
                },
                status=400,
            )
        user = User.objects.filter(email=forgotten_password.first().email)
        if not user.exists():
            return Response(
                {"status_code": 400, "errors": [{"message": "an error occurred"}]},
                status=400,
            )
        user_to_change = user.first()
        user_to_change.set_password(self.request.data["password"])
        user_to_change.save()
        return Response(
            {"status_code": 200, "message": "your password has been reset!"},
            status=200,
        )



# Generated by Django 2.2.13 on 2020-11-05 11:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0030_auto_20201103_1718"),
    ]

    operations = [
        migrations.AddField(
            model_name="task",
            name="correct",
            field=models.BooleanField(null=True),
        ),
    ]



from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization
from human_lambdas.workflow_handler.models import WebHook
from human_lambdas.workflow_handler.tests.constants import (
    REGISTRATION_DATA,
    WORKFLOW_DATA_3,
)


class TestZapierHook(APITestCase):
    def setUp(self):

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]

        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get("/v1/users/api-token")
        self.token = response.data["token"]

        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            WORKFLOW_DATA_3,
            format="json",
        )
        self.workflow_id = response.data["id"]

    def test_zapier_authentication(self):
        self.client.credentials(HTTP_AUTHORIZATION="Token " + self.token)
        response = self.client.get("/zapier/authentication")
        self.assertTrue("organization" in response.data)

    def test_subscribe_to_hook(self):
        self.client.credentials(HTTP_AUTHORIZATION="Token " + self.token)
        hook_url = "https://some.url.se"
        response = self.client.post(
            "/zapier/hook",
            {
                "target": hook_url,
                "workflow_id": self.workflow_id,
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        webhook = WebHook.objects.filter(pk=response.data["id"])
        self.assertTrue(webhook.exists())
        self.assertEqual(webhook.first().target, hook_url)
        self.assertTrue(webhook.first().is_zapier)

    def test_unsubscribe_to_hook(self):
        self.client.credentials(HTTP_AUTHORIZATION="Token " + self.token)
        hook_url = "https://some.url.se"
        response = self.client.post(
            "/zapier/hook",
            {
                "target": hook_url,
                "workflow_id": self.workflow_id,
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        hook_id = response.data["id"]
        webhook = WebHook.objects.filter(pk=hook_id)
        self.assertTrue(webhook.exists())

        response = self.client.delete(
            "/zapier/hook",
            {"hook_id": hook_id},
            format="json",
        )
        self.assertEqual(
            response.status_code, status.HTTP_204_NO_CONTENT, response.data
        )
        webhook = WebHook.objects.filter(pk=hook_id)
        self.assertFalse(webhook.exists())



import json
import logging
import pathlib

from django.db import models
from rest_framework import serializers
from rest_framework.generics import ListAPIView

from human_lambdas.user_handler.models import User

logger = logging.getLogger(__name__)


class Template(models.Model):
    name = models.TextField()
    summary = models.TextField()
    thumbnail = models.TextField()


class TemplatesSerializer(serializers.ModelSerializer):
    class Meta:
        model = Template
        fields = ["name", "summary", "thumbnail", "id"]


class TemplatesView(ListAPIView):
    queryset = User.objects.all()
    serializer_class = TemplatesSerializer

    def get_queryset(self):
        try:
            text = pathlib.Path("templates.json").read_text()
            return map(lambda t: Template(**t), json.loads(text))
        except Exception:
            return []



from typing import Any, List, Tuple

from django.db.models import Q
from django.shortcuts import get_object_or_404
from rest_framework import status
from rest_framework.generics import CreateAPIView
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response

from human_lambdas.data_handler.data_transformation import transform_ext2int
from human_lambdas.external.authentication import TokenAuthentication
from human_lambdas.hl_rest_api import analytics
from human_lambdas.user_handler.notifications import send_notification
from human_lambdas.user_handler.permissions import IsOrgAdmin
from human_lambdas.workflow_handler.audits import GetCompletedTaskView
from human_lambdas.workflow_handler.models import Task, Workflow
from human_lambdas.workflow_handler.utils import notify_slack, parse_dates

from .serializers import ExternalTaskSerializer


class GetExternalCompletedTaskView(GetCompletedTaskView):
    """
    External API View for getting all the Tasks
    """

    permission_classes = (IsAuthenticated, IsOrgAdmin)
    authentication_classes = (TokenAuthentication,)
    serializer_class = ExternalTaskSerializer

    def get_queryset(self, *args, **kwargs):
        user = self.request.user
        workflows = Workflow.objects.filter(
            Q(organization__pk=self.request.auth.organization_id)
            & Q(disabled=False)
            & Q(organization__pk=self.kwargs["org_id"])
        )
        workflow = get_object_or_404(
            workflows, pk=self.kwargs.get("queue_id", self.kwargs.get("workflow_id"))
        )
        analytics.track(
            user.pk,
            "Get Completed Tasks",
            {
                "workflow_id": self.kwargs.get(
                    "queue_id", self.kwargs.get("workflow_id")
                )
            },
        )
        return (
            Task.objects.defer("data")
            .filter(
                Q(workflow=workflow)
                & Q(status="completed")
                & Q(completed_at__range=(parse_dates(self.request)))
            )
            .filter(*args, **kwargs)
            .order_by("-completed_at")
        )


def get_data_value(request_data, w_data):
    input_value = request_data[w_data["id"]]
    if w_data["type"] == "list" and w_data[w_data["type"]]["subtype"] == "number":
        input_value = [float(i) for i in input_value]
    return input_value


class CreateTaskView(CreateAPIView):
    """
    External API View for creating Tasks
    """

    authentication_classes = (TokenAuthentication,)
    permission_classes = (IsAuthenticated, IsOrgAdmin)
    serializer_class = ExternalTaskSerializer

    def get_queryset(self):
        workflows = Workflow.objects.filter(
            Q(organization__pk=self.request.auth.organization_id)
            & Q(organization__pk=self.kwargs["org_id"] & Q(disabled=False))
        )
        return Task.objects.defer("data").filter(
            Q(workflow__in=workflows)
            & Q(
                workflow__id=self.kwargs.get("queue_id", self.kwargs.get("workflow_id"))
            )
        )

    def post(self, request: Request, *args, **kwargs):
        if "data" not in self.request.data or not self.request.data["data"]:
            return Response(
                {"status_code": 400, "errors": [{"message": "No data"}]},
                status=400,
            )
        return self.create(request, *args, **kwargs)

    def create(self, request: Request, *args, **kwargs):
        formatted_data, workflow = self.preprocess_data()

        if workflow.is_running:
            notify_slack(f"API Task created for {workflow.name}", request)

        serializer = self.get_serializer(data={**request.data, "data": formatted_data})
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        send_notification(workflow)
        headers = self.get_success_headers(serializer.data)
        return Response(
            serializer.data, status=status.HTTP_201_CREATED, headers=headers
        )

    def preprocess_data(self) -> Tuple[List[Any], Workflow]:
        workflow: Workflow = get_object_or_404(
            Workflow, pk=self.kwargs.get("queue_id", self.kwargs.get("workflow_id"))
        )
        formatted_data = transform_ext2int(workflow.data, self.request.data["data"])

        return formatted_data, workflow

    def perform_create(self, serializer):
        serializer.save(source_name="api")



import logging

from django.conf import settings
from django.shortcuts import get_object_or_404
from django.utils import timezone
from rest_framework.permissions import AllowAny, IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from human_lambdas.hl_rest_api import analytics
from human_lambdas.hl_rest_api.utils import (
    generate_unique_token,
    is_invalid_email,
)
from human_lambdas.user_handler.models import (
    Invitation,
    Notification,
    Organization,
    User,
)
from human_lambdas.user_handler.permissions import IsOrgAdmin
from human_lambdas.workflow_handler.models import WorkflowNotification

from .apps import UserHandlerConfig

logger = logging.getLogger(__name__)


class SendInviteView(APIView):
    permission_classes = (IsAuthenticated, IsOrgAdmin)

    def get_queryset(self):
        return Invitation.objects.filter(
            organization__pk=self.kwargs["org_id"]
        ).distinct()

    def get(self, request, *args, **kwargs):
        invited_users = [
            {
                "email": invitation.email,
                "pending": True,
                "is_admin": invitation.admin,
                "link": invitation.invite_link,
            }
            for invitation in self.get_queryset()
        ]
        return Response(
            {"status_code": 200, "invited_users": invited_users}, status=200
        )

    def post(self, request, *args, **kwargs):
        email_set = set("".join(request.data["emails"].split()).split(","))
        invalid_email_list, already_added_email_list = [], []
        organization = Organization.objects.get(pk=kwargs["org_id"])
        template_data = []
        emails = []
        links = {}
        for email in email_set:
            if is_invalid_email(email):
                invalid_email_list.append(email)
                continue

            if organization.user.filter(email=email).exists():
                already_added_email_list.append(email)
                continue

            for inv in self.get_queryset().filter(email=email):
                inv.delete()

            token = generate_unique_token(email, kwargs["org_id"])

            if not User.objects.filter(email=email).exists():
                invite_link = "{0}invite/{1}".format(settings.FRONT_END_BASE_URL, token)
                analytics.track(self.request.user.pk, "Invite user", {"new_user": True})
                analytics.track(email, "Invited new user")
            else:
                invite_link = "{0}invite/success/{1}".format(
                    settings.FRONT_END_BASE_URL, token
                )
                analytics.track(
                    self.request.user.pk, "Invite user", {"new_user": False}
                )

            template_data.append(
                {
                    "org_name": organization.name,
                    "inviter_name": request.user.name,
                    "url": invite_link,
                }
            )

            Invitation(
                email=email,
                organization=organization,
                invited_by=request.user,
                token=token,
                invite_link=invite_link,
                expires_at=timezone.now()
                + timezone.timedelta(days=settings.INVITATION_EXPIRATION_WINDOW_DAYS),
            ).save()

            links[email] = invite_link
            emails.append(email)

        if settings.INVITATION_TEMPLATE is None or settings.ACCOUNT_ASM_GROUPID is None:
            logger.warn("Sendgrid not set up, ignoring email trigger")
        else:
            UserHandlerConfig.emailclient.send_email(
                to_email=emails,
                template_id=settings.INVITATION_TEMPLATE,
                template_data=template_data,
                group_id=int(settings.ACCOUNT_ASM_GROUPID),
            )

        # sending responses
        if len(invalid_email_list) == 0 and len(already_added_email_list) == 0:
            return Response(
                {
                    "status_code": 200,
                    "message": "all emails were successfully added!",
                    "links": links,
                },
                status=200,
            )
        response_text = ""
        for email in invalid_email_list:
            response_text += "{0} is an invalid email. ".format(email)
        for email in already_added_email_list:
            response_text += "{0} is already a part of the organization. ".format(email)
        return Response(
            {
                "status_code": 400,
                "errors": [{"message": response_text}],
                "links": links,
            },
            status=400,
        )

    def patch(self, request, *args, **kwargs):
        invites = self.get_queryset().filter(email=self.request.data["email"])
        if not invites.exists():
            return Response(
                {
                    "status_code": 400,
                    "errors": [
                        {"message": "there are no invitations to this user to update"}
                    ],
                },
                status=400,
            )
        for invite in invites:
            invite.admin = True if self.request.data["admin"] else False
            invite.save()
        if self.request.data["admin"]:
            response_text = "this invitation has now been set to admin status"
        else:
            response_text = "this invitation has now been set to worker status"
        return Response({"status_code": 200, "message": response_text}, status=200)

    def delete(self, request, *args, **kwargs):
        invites = self.get_queryset().filter(email=self.request.data["email"])
        if not invites.exists():
            return Response(
                {
                    "status_code": 400,
                    "errors": [
                        {"message": "there are no invitations to this user to delete"}
                    ],
                },
                status=400,
            )
        for invite in invites:
            invite.delete()
        response_text = "this invite has now been deleted"
        return Response({"status_code": 200, "message": response_text}, status=200)


class InvitationView(APIView):
    permission_classes = (AllowAny,)

    def get_queryset(self):
        return Invitation.objects.filter(token=self.kwargs["invite_token"])

    def get(self, request, *args, **kwargs):
        invite = get_object_or_404(self.get_queryset())
        invitation_email, invitation_org = (
            invite.email,
            invite.organization.name,
        )
        analytics.track(invite.email, "Invited new user visit")
        return Response(
            {
                "status_code": 200,
                "invitation_email": invitation_email,
                "invitation_org": invitation_org,
            },
            status=200,
        )

    def post(self, request, *args, **kwargs):
        invite = get_object_or_404(self.get_queryset())
        if invite.expires_at < timezone.now():
            return Response(
                {
                    "status_code": 400,
                    "errors": [{"message": "this token has expired!"}],
                },
                status=400,
            )

        invitation_org = invite.organization
        if invitation_org.user.filter(email=invite.email).exists():
            return Response(
                {
                    "status_code": 400,
                    "errors": [
                        {"message": "this organization has already been joined"}
                    ],
                },
                status=400,
            )
        user = User.objects.filter(email=invite.email).first()
        if not user:
            notification = Notification()
            notification.save()
            user = User(
                email=invite.email,
                name=request.data["name"],
                current_organization_id=invitation_org.id,
                notifications=notification,
            )
            user.set_password(request.data["password"])
            user.save()
            response_data = {
                "data": {
                    "status_code": 201,
                    "message": "Your account has been created!",
                    "email": invite.email,
                },
                "status": 201,
            }
            analytics.track(user.email, "Invited new user signed up")
            analytics.alias(user.email, user.pk)
            analytics.track(user.pk, "User Signup", {"source": "invite"})
        else:
            response_data = {
                "data": {
                    "status_code": 200,
                    "message": "Success!",
                    "email": invite.email,
                },
                "status": 200,
            }
            analytics.track(user.pk, "Existing user joined new org", {"new": False})

        for workflow in invitation_org.workflow_set.all():
            WorkflowNotification(
                workflow=workflow, notification=user.notifications
            ).save()

        if invite.admin:
            invitation_org.add_admin(user)
            analytics.identify(
                user.pk,
                {
                    "is_admin": True,
                    "org_id": invitation_org.pk,
                    "org_name": invitation_org.name,
                    "name": user.name,
                    "email": user.email,
                    "user_id": user.pk,
                },
            )
        else:
            invitation_org.user.add(user)
            analytics.identify(
                user.pk,
                {
                    "is_admin": False,
                    "org_id": invitation_org.pk,
                    "org_name": invitation_org.name,
                    "name": user.name,
                    "email": user.email,
                    "user_id": user.pk,
                },
            )

        Invitation.objects.filter(
            organization__pk=invitation_org.id, email=invite.email
        ).all().delete()

        return Response(**response_data)



from unittest.mock import patch

import pytest
from django.test import TestCase
from parameterized import parameterized

from human_lambdas.user_handler.models import Organization, User
from human_lambdas.workflow_handler.models import Task, Workflow
from human_lambdas.workflow_handler.region import Region

DB_DATA = {"foo": 2}
BUCKET_DATA = {"blah": 3}
TASK_PK = 1


class TestR13n(TestCase):
    def setUp(self):
        user = User(name="foo", email="foo@bar.com")
        user.save()
        self.user = user
        organization = Organization(name="fooInc")
        organization.save()
        organization.add_admin(user)
        self.workflow = Workflow(
            name="lkj",
            inputs={},
            outputs={},
            created_by=user,
            organization=organization,
        )
        self.workflow.save()

    @parameterized.expand([(Region.US,), (Region.AU,)])
    def test_when_non_eu_data_accessed_then_pulled_from_bucket(self, region: Region):
        task = Task(
            pk=TASK_PK,
            workflow=self.workflow,
            inputs={},
            outputs={},
            data=DB_DATA,
            region=region.name,
        )

        with patch(
            "human_lambdas.workflow_handler.regional_storage.retrieve"
        ) as retrieve, patch(
            "human_lambdas.workflow_handler.regional_storage.store"
        ) as store:
            task.save()

            retrieve.return_value = BUCKET_DATA
            t = Task.objects.get(pk=TASK_PK)

            data = t.data

            assert data == BUCKET_DATA
            calls = retrieve.mock_calls

            assert len(calls) == 1
            assert calls[0].args == (1, region)

    def test_when_non_eu_data_accessed_but_field_deferred_then_no_data(self):
        task = Task(
            pk=TASK_PK,
            workflow=self.workflow,
            inputs={},
            outputs={},
            data=DB_DATA,
            region=Region.AU.name,
        )

        with patch(
            "human_lambdas.workflow_handler.regional_storage.retrieve"
        ) as retrieve, patch(
            "human_lambdas.workflow_handler.regional_storage.store"
        ) as store:
            task.save()

            retrieve.return_value = BUCKET_DATA
            t = Task.objects.defer("data").get(pk=TASK_PK)

            assert len(retrieve.mock_calls) == 0

    def test_when_eu_data_accessed_then_pulled_from_db(self):
        region = None
        task = Task(
            pk=TASK_PK,
            workflow=self.workflow,
            inputs={},
            outputs={},
            data=DB_DATA,
            region=region.name if region else None,
        )
        task.save()

        with patch(
            "human_lambdas.workflow_handler.regional_storage.retrieve"
        ) as retrieve:
            retrieve.return_value = BUCKET_DATA
            t = Task.objects.get(pk=TASK_PK)

            data = t.data

            assert data == DB_DATA
            assert len(retrieve.mock_calls) == 0

    def test_when_eu_data_saved_then_stored_in_db(self):
        region = None
        task = Task(
            pk=TASK_PK,
            workflow=self.workflow,
            inputs={},
            outputs={},
            data=DB_DATA,
            region=region,
        )
        with patch(
            "human_lambdas.workflow_handler.regional_storage.store"
        ) as store, patch("django.db.models.Model.save", autospec=True) as save:
            task.save()

            assert len(store.mock_calls) == 0

            assert save.call_count == 1
            assert save.call_args[0][0].data == DB_DATA

    def test_when_non_eu_data_no_id_saved_then_stored_in_bucket(self):
        region = Region.AU

        task = Task(
            workflow=self.workflow,
            inputs={},
            outputs={},
            data=BUCKET_DATA,
            region=region.name,
        )
        with patch(
            "human_lambdas.workflow_handler.regional_storage.store"
        ) as store, patch("django.db.models.Model.save", autospec=True) as save:

            def set_id(self: Task):
                self.pk = TASK_PK

            save.side_effect = set_id
            store.return_value = BUCKET_DATA
            task.save()

            assert len(store.mock_calls) == 1
            assert store.mock_calls[0].args == (1, region, BUCKET_DATA)

            assert save.call_count == 1

    @pytest.mark.bucket
    def test_when_data_not_fetched_then_no_regional_update(self):
        region = Region.AU

        Task(
            pk=TASK_PK,
            workflow=self.workflow,
            inputs={},
            outputs={},
            data=BUCKET_DATA,
            region=region.name,
        ).save()

        task = Task.objects.defer("data").get(pk=TASK_PK)

        with patch(
            "human_lambdas.workflow_handler.regional_storage.store"
        ) as store, patch(
            "human_lambdas.workflow_handler.regional_storage.retrieve"
        ) as retrieve:
            task.save()
            assert len(store.mock_calls) == 0
            assert len(retrieve.mock_calls) == 0



# Generated by Django 2.2.5 on 2019-12-16 18:41

from django.db import migrations, models

import human_lambdas.workflow_handler.fields


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0003_auto_20191119_1457"),
    ]

    operations = [
        migrations.AddField(
            model_name="task",
            name="created_at",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name="task",
            name="output_data",
            field=human_lambdas.workflow_handler.fields.JSONField(null=True),
        ),
    ]



from typing import List, Optional, cast

import pytest

from human_lambdas.data_handler.data_validation import (
    DataValidationError,
    data_validation,
)
from human_lambdas.data_handler.tests.factories import (
    DateBlockFactory,
    EmailBlockFactory,
    EmailFormBlockFactory,
    make_block,
)

VALID_URL = "http://localhost.com/workflows/10/edit.png"
INVALID_URL = "http:/"


def get_url_test():
    for field in ["value", "placeholder"]:
        for url in cast(List[Optional[str]], [VALID_URL, INVALID_URL, None, ""]):
            for type in [
                "embed",
                "audio",
                "image",
                "video",
                "pdf",
                "link",
                "bounding_boxes",
            ]:
                yield (field, url, type)


class TestValidators:
    def test_when_invalid_email_then_fail(self):
        block = make_block(FACTORY_CLASS=EmailBlockFactory, email__value="bla")
        try:
            data_validation([block])
            assert False
        except DataValidationError:
            pass

    def test_when_no_email_then_pass(self):
        block = make_block(FACTORY_CLASS=EmailBlockFactory)
        data_validation([block])

    def test_when_no_placeholder_then_pass(self):
        block = make_block(
            FACTORY_CLASS=EmailBlockFactory,
            email={"read_only": False, "use_placeholder": False},
        )
        data_validation([block])

    def test_when_invalid_placeholder_then_fail(self):
        try:
            block = make_block(
                FACTORY_CLASS=EmailBlockFactory, email__placeholder="bla"
            )
            data_validation([block])
            assert False
        except DataValidationError:
            pass

    def test_when_email_empty_then_pass(self):
        block = make_block(FACTORY_CLASS=EmailBlockFactory, email__value="")
        data_validation([block])

    def test_when_email_form_then_pass(self):
        block = make_block(
            FACTORY_CLASS=EmailFormBlockFactory,
            form_sequence__data__0__email__value="blah@b.co",
        )
        data_validation([block])

    def test_when_no_email_form_then_pass(self):
        block = make_block(
            FACTORY_CLASS=EmailFormBlockFactory,
            form_sequence__data__0__email__value=None,
        )
        data_validation([block])

    def test_when_empty_email_form_then_pass(self):
        block = make_block(
            FACTORY_CLASS=EmailFormBlockFactory, form_sequence__data__0__email__value=""
        )
        data_validation([block])

    def test_when_invalid_email_form_then_fail(self):
        try:
            block = make_block(
                FACTORY_CLASS=EmailFormBlockFactory,
                form_sequence__data__0__email__value="blah",
            )
            data_validation([block])
            assert False
        except DataValidationError:
            pass

    @pytest.mark.parametrize("field, url, type", get_url_test())
    def test_url_validation(self, field, url, type):
        block_opts = {type: {field: url}, "type": type}

        if type == "bounding_boxes":
            block_opts["bounding_boxes"]["value"] = (
                {"image": url} if field == "value" else {}
            )

        block = make_block(**block_opts)

        if url == INVALID_URL:
            with pytest.raises(DataValidationError):
                data_validation([block])
        else:
            data_validation([block])

    def test_when_empty_bounding_boxes_workflow_then_pass(self):
        bounding_boxes = {
            "options": [
                {"id": "option_1", "name": "Option 1"},
                {"id": "option_2", "name": "Option 2"},
            ],
            "placeholder": "",
        }
        block = make_block(
            bounding_boxes=bounding_boxes,
            type="bounding_boxes",
        )

        data_validation([block], is_workflow=True)

    def test_when_date_block_empty_workflow_then_pass(self):
        block = make_block(FACTORY_CLASS=DateBlockFactory)
        data_validation([block], is_workflow=True)

    def test_when_date_block_empty_task_then_pass(self):
        block = make_block(FACTORY_CLASS=DateBlockFactory)
        data_validation([block], is_workflow=False)

    def test_when_date_block_valid_then_pass(self):
        block = make_block(
            FACTORY_CLASS=DateBlockFactory,
            date__value="2020-01-31",
            date__placeholder="2020-01-31",
        )
        data_validation([block])

    def test_when_date_block_invalid_then_fail(self):
        block = make_block(
            FACTORY_CLASS=DateBlockFactory,
            date__value="2020/01/31",  # Invalid according to ISO-8601 (YYYY-MM-DD)
        )
        try:
            data_validation([block])
            assert False
        except:
            pass

        block = make_block(
            FACTORY_CLASS=DateBlockFactory,
            date__value="2020-31-01",  # Invalid according to ISO-8601 (YYYY-MM-DD)
        )
        try:
            data_validation([block])
            assert False
        except:
            pass

    def test_when_date_block_has_no_content_then_pass(self):
        block = make_block(
            FACTORY_CLASS=DateBlockFactory, date={"read_only": True}, type="date"
        )
        data_validation([block])
